'''A class for computing the
parameters for Marcus theory
from two constrained DFT
wave functions

Computes:
-the coupling matrix Hab
 Hab = <Psi_a|H|Psi_b>
-reorganization energy lambda
lambda = E_a(Rb)-E_a(Ra)
'''

import numpy as np
from gpaw.cdft.cdft import (WeightFunc, get_ks_energy_wo_external,
                            get_all_weight_functions)
from ase.units import kB as kb
from gpaw.utilities.ps2ae import PS2AE, interpolate_weight
from ase.units import Bohr
from gpaw.mpi import rank
import warnings


spin_state_error = ('The cDFT wave functions have\n' +
                    'different spin states! Similar\n' +
                    'spin states are required for coupling constant\n' +
                    'calculation!')

ae_ibz_error = ('The all electron calculation is unreliable with kpts\n' +
                'Please set AE to false')

nab_missing_error = ('The pair density matrix n_ab must\n' +
                     'be provided for weight matrix calculation')
migliore_warning = ('WARNING! Migliore coupling might be unreliable!:\n' +
                    '<F|GS> =<I|GS> and dE_if=0')


class CouplingParameters:
    def __init__(self,
                 cdft_a=None,
                 cdft_b=None,
                 h=0.04,
                 calc_a=None,
                 calc_b=None,
                 calc_gs=None,
                 wfs_a='initial.gpw',
                 wfs_b='final.gpw',
                 AE=False,
                 charge_difference=False,
                 charge_regions_A=None,
                 spin_regions_A=None,
                 charge_regions_B=None,
                 spin_regions_B=None,
                 FA=None,
                 FB=None,
                 Va=None,
                 Vb=None,
                 NA=None,
                 NB=None,
                 n_ab=None,
                 VW_ab=None,
                 VW_ba=None,
                 w_k=None,
                 Rc={},
                 mu={},
                 specific_bands_A=None,
                 specific_bands_B=None,
                 kpoint=None,
                 spin=None,
                 S_matrix='S_matrix',
                 VW_matrix='VW_matrix',
                 S=None,
                 A=None,
                 B=None,
                 save_matrix=False,
                 freq=None,
                 temp=None,
                 E_tst=None,
                 reorg=None,
                 use_adiabatic_correction=False,
                 reaction_energy=0.,
                 band_occupation_cutoff=0.5,
                 energy_gap=None):
        '''cdft_a cdft_b: cdft calculators
        h = grid spacing in S_AB and W_AB calculations in AE mode
        AE = Use all electron wave functions


        if cdft calculators are not provided,
        calc_a, calc_b, wfs_a, wfs_b,gd,
        FA,FB,Va, Vb, NA, NB,
        charge_regions_A/B must be provided

        calc_a/b = GPAW calculators for a/b states
        wfs_a/b = wfs of the a/b states
        FA/FB = pseudo cdft free energies of A/B states in eV
                as from cdft.free_energy() or cdft.Ecdft

        Va/b = list of lagrange multipliers in eV
        NA/NB = charge + spin constraints in A/B states
             !NOTE! Total number of electrons as from cdft.get_constraints()

        weight_A/B = the weight functions for A/B
            Array

        N_charge_regions_A/B = number of constrained charges
            int

        E_KS_A/B = KS energies with out the external potential
                  (Ekin+Epot+Ebar+Exc+S) in eV

        n_ab = S_AB_ij(k) = <A_i(k)|B_j(k)>

        VW_ab,VW_ba = <A(k)|sum_c Vc_B*Wc_B|B(k)>, <B(k)|sum_c Vc_A*Wc_A|B(k)>

        WAB, WBA = sum_k sum_ij sum_c <A_i(k)|Wc_B|B_j(k)>

        w_k = combined kpt weight, (w_kA*w_kB) in eq 48 and 49
              can be provided to compute overlap and
              weight matrices

        specific_bands, kpoint, spin = [[],[]], int, int
            allows to extract density matrix elements
            and weight-density matrix elements for
            specified orbitals Eq. 48 and 49
            [[],[]] = indices of first for alpha,
                   second for beta spin bands

            If one these is activated, matrix of
            int[psi_sk^i*psi_sk^j] is returned instead
            of S_ab or W_ab

        S_matrix, W_matrix = str
            place to print density and
            weight-density matrix
            Turned on automatically if bands,
            spin, or kpoint are specified

        S, A, B = np.array, stored migliore matrices

        for calculation of electron transfer rates
        using the Marcus theory the following can be
        specified:

        freq = effective frequency
        temp = temperature
        E_tst = adiabatic reaction barrier
        use_adiabatic_correction = False/True
            add the adiabadicity correction to
            Marcus barrier (can some times give very
            too small barriers for an adiabatic reaction
            with large coupling...)
        reaction_energy: E_B(RB)-E_A(RA)
        band_occupation_cutoff = minimum filling of band
            to be included in the wfs
        Rc, mu: Gaussian parameters. If not provided either standard or
            the ones from a cdft will be used
        '''

        self.charge_difference = charge_difference
        self.AE = AE

        if cdft_a is not None and cdft_b is not None:
            self.cdft_A = cdft_a
            self.cdft_B = cdft_b
            self.Rc = self.cdft_A.Rc
            self.mu = self.cdft_A.mu

            self.calc_A = self.cdft_A.calc
            self.calc_B = self.cdft_B.calc

            if self.AE:
                self.gd = self.cdft_A.get_grid()
            else:
                self.gd = self.cdft_A.calc.density.gd

            # cDFT free energies
            self.FA = self.cdft_A.cdft_free_energy()
            self.FB = self.cdft_B.cdft_free_energy()

            # cDFT energies
            self.EA = self.cdft_A.dft_energy()
            self.EB = self.cdft_B.dft_energy()

            # lagrange multipliers
            self.Va = self.cdft_A.get_lagrangians()
            self.Vb = self.cdft_B.get_lagrangians()

            # constraint values
            self.NA = self.cdft_A.get_constraints()
            self.NB = self.cdft_B.get_constraints()

            # number of charge regions
            self.n_charge_regionsA = self.cdft_A.n_charge_regions
            self.n_charge_regionsB = self.cdft_B.n_charge_regions
            self.regionsA = self.cdft_A.regions
            self.regionsB = self.cdft_B.regions
            self.atoms = self.calc_A.get_atoms()

            if self.AE:
                self.fineweightA = self.cdft_A.get_weight(save=False, pad=True)
                self.fineweightB = self.cdft_B.get_weight(save=False, pad=True)
            else:
                self.coarseweightA = get_all_weight_functions(
                    self.calc_A.atoms, self.gd, self.regionsA,
                    self.charge_difference, self.Rc, self.mu)
                self.coarseweightB = get_all_weight_functions(
                    self.calc_B.atoms, self.gd, self.regionsB,
                    self.charge_difference, self.Rc, self.mu)

        else:
            self.calc_A = calc_a
            self.calc_B = calc_b

            if self.AE:
                self.gd = self.calc_A.density.finegd
            else:
                self.gd = self.calc_A.density.gd

            # cDFT free energies
            self.FA = FA
            self.FB = FB

            # Weights, i.e., lagrange multipliers
            self.Va = Va
            self.Vb = Vb
            regionsA = []
            if charge_regions_A is not None:
                regionsA += charge_regions_A
            if spin_regions_A is not None:
                regionsA += spin_regions_A

            regionsB = []
            if charge_regions_B is not None:
                regionsB += charge_regions_B
            if spin_regions_B is not None:
                regionsB += spin_regions_B

            self.regionsA = regionsA
            self.regionsB = regionsB
            self.NA = NA
            self.NB = NB

            # see if a correct set of Rc and mu are provided
            if (set(Rc) == set(self.calc_A.atoms.get_chemical_symbols()) and
                    set(mu) == set(self.calc_A.atoms.get_chemical_symbols())):
                self.Rc = Rc
                self.mu = mu
            else:
                # get mu and Rc the long way
                w_temp = WeightFunc(gd=self.gd,
                                    atoms=self.calc_A.atoms,
                                    indices=self.regionsA,
                                    Rc=Rc,
                                    mu=mu)
                self.Rc, self.mu = w_temp.get_Rc_and_mu()
                del w_temp

            if self.AE:
                self.fineweightA = get_all_weight_functions(
                    self.calc_A.atoms, self.gd, self.regionsA,
                    self.charge_difference, self.Rc, self.mu)

                self.fineweightB = get_all_weight_functions(
                    self.calc_B.atoms, self.gd, self.regionsB,
                    self.charge_difference, self.Rc, self.mu)

            else:
                self.coarseweightA = get_all_weight_functions(
                    self.calc_A.atoms, self.gd, self.regionsA,
                    self.charge_difference, self.Rc, self.mu)
                self.coarseweightB = get_all_weight_functions(
                    self.calc_B.atoms, self.gd, self.regionsB,
                    self.charge_difference, self.Rc, self.mu)

            self.n_charge_regionsA = len(charge_regions_A)
            self.n_charge_regionsB = len(charge_regions_B)

            # Do not include external potential
            self.EA = get_ks_energy_wo_external(self.calc_A)
            self.EB = get_ks_energy_wo_external(self.calc_B)

        self.finegd = self.calc_A.density.finegd
        self.energy_gap = energy_gap
        self.Va = np.asarray(self.Va)
        self.Vb = np.asarray(self.Vb)

        # wave functions
        if wfs_a is not None and wfs_b is not None:
            self.wfs_A = wfs_a
            self.wfs_B = wfs_b
        else:
            self.wfs_A = self.calc_A.wfs
            self.wfs_B = self.calc_B.wfs

        # ground state calculator for migliore

        if calc_gs is not None:
            self.calc_gs = calc_gs
        if S is not None:
            self.S = np.load(S)
        if A is not None:
            self.A = np.load(A)
        if B is not None:
            self.B = np.load(B)
        if w_k is not None:
            self.w_AB = self.w_AC = self.w_BC = np.load(w_k)

        # use precalculated pair density/weight matrices
        if w_k:
            self.w_k = np.load(w_k)
        if n_ab:
            self.n_ab = np.load(n_ab)

        if VW_ab:
            self.VW_AB = np.load(VW_ab)
            self.VW_BA = np.load(VW_ba)

        # specify bands to be treated
        self.specific_bands_A = specific_bands_A
        self.specific_bands_B = specific_bands_B

        if kpoint:
            self.kpoints = kpoint
        else:
            self.kpoints = self.calc_A.wfs.kd.nibzkpts
        self.spin = spin

        # where to save matrices
        self.S_matrix = S_matrix
        self.VW_matrix = VW_matrix

        self.save_matrix = save_matrix
        # Only part of bands treated --> matrices saved
        if self.specific_bands_A or self.specific_bands_B or self.spin:
            self.save_matrix = True

        self.H_AA = self.EA
        self.H_BB = self.EB

        self.density = self.calc_A.density.finegd.zeros()
        self.atoms = self.calc_A.atoms
        self.natoms = len(self.atoms)
        self.h = h  # all-electron interpolation grid spacing

        # controls for rate calculation

        self.E_tst = E_tst  # adiabatic barrier

        # effective frequency
        if freq:
            self.freq = freq
        else:
            self.freq = 1.e12
        # temperature
        if temp:
            self.temp = temp
        else:
            self.temp = 298
        self.reorg = reorg

        self.reaction_energy = reaction_energy
        self.use_adiabatic_correction = use_adiabatic_correction
        self.band_occupation_cutoff = band_occupation_cutoff

    def get_coupling_term(self):
        # coupling by Migliore
        # dx.doi.org/10.1021/ct200192d, eq 6
        # check that needed terms exist

        if (hasattr(self, 'H') and hasattr(self, 'S')):
            if rank == 0:
                H = self.H.copy()
                S = self.S.copy()

        else:
            self.make_hamiltonian_matrix()
            if rank == 0:
                H = self.H.copy()
                S = self.S.copy()
        if rank == 0:
            H_av = 0.5 * (H[0][1] + H[1][0])
            S_av = 0.5 * (S[0][1] + S[1][0])

            self.ct = (1. / (1 - S_av**2)) * np.abs(H_av - S_av *
                                                    (H[0][0] + H[1][1]) / 2.)
            return self.ct

    def get_coupling_term_from_lowdin(self):
        self.make_hamiltonian_matrix()
        if rank == 0:
            H_orth = self.lowdin_orthogonalize_cdft_hamiltonian()
            self.ct = 1. / 2. * np.real(H_orth[0][1] + H_orth[1][0])
            return self.ct

    def lowdin_orthogonalize_cdft_hamiltonian(self):
        # H_orth = S^(-1/2)*H_cdft*S^(-1/2)

        U, D, V = np.linalg.svd(self.S, full_matrices=True)  # V=U(+)
        s = np.diag(D**(-0.5))
        S_square = np.dot(U, np.dot(s, V))  # S^(-1/2)

        H_orth = np.dot(S_square, np.dot(self.H, S_square))
        return H_orth

    def get_migliore_coupling(self):
        # the entire migliore method
        # from dx.doi.org/10.1021/ct200192d,
        # requires also ground state adiabatic wf
        if not hasattr(self, 'S') or not hasattr(self, 'A') or not hasattr(
                self, 'B'):
            self.S, self.A, self.B, self.w_AB, self.w_AC, self.w_BC = (
                self.get_migliore_wf_overlaps(
                    self.calc_A, self.calc_B, self.calc_gs))
        if rank == 0:
            SIF = 0.
            SFI = 0.
            A = 0.
            B = 0.

            for k in range(len(self.w_AB)):
                SIF += self.w_AB[k] * np.linalg.det(self.S[k])
                SFI += self.w_AB[k] * np.linalg.det(
                    np.transpose(np.conjugate(self.S[k])))
            for k in range(len(self.w_AC)):
                A += self.w_AC[k] * np.linalg.det(self.A[k])
            for k in range(len(self.w_BC)):
                B += self.w_BC[k] * np.linalg.det(self.B[k])

            if self.energy_gap is None:
                self.get_energy_gap()
            dE_if = self.energy_gap
            if np.abs(dE_if) < 0.001 and np.abs(A**2 - B**2) < 0.001:
                warnings.warn(migliore_warning)
            self.ct = np.abs((A * B / (A**2 - B**2)) * dE_if *
                             (1 - SIF * (A**2 + B**2) / (2 * A * B)) *
                             1 / (1 - SIF**2))
            return self.ct

    def make_hamiltonian_matrix(self):
        # this returns a 2x2 cDFT Hamiltonian
        # in a non-orthogonal basis -->
        # made orthogonal in get_coupling_term
        # self.get_diagonal_H_elements()
        H_AA = self.H_AA  # diabat A with H^KS_A
        H_BB = self.H_BB  # diabat B with H^KS_B

        if hasattr(self, 'S'):
            S = self.S
        else:
            S = self.get_overlap_matrix()
            self.S = S

        if hasattr(self, 'VW'):
            VW = self.VW
        else:
            VW = self.get_weight_matrix()
        if rank == 0:
            S_AB = S[0][1]
            S_BA = S[1][0]

            h_AB = self.FB * S_AB - VW[0][1]
            h_BA = self.FA * S_BA - VW[1][0]

            # Ensure that H is hermitian
            H_AB = 1. / 2. * (h_AB + h_BA)
            H_BA = 1. / 2. * (h_AB + h_BA).conj()

            self.H = np.array([[H_AA, H_BA], [H_AB, H_BB]])
            return self.H

    def get_ae_pair_weight_matrix(self):
        if self.calc_A.wfs.kd.nibzkpts != 1:
            raise ValueError(ae_ibz_error)
        if hasattr(self, 'n_ab'):
            pass
        else:
            raise ValueError(nab_missing_error)

        # pseudo wfs to all-electron wfs
        psi_A = PS2AE(self.calc_A, grid_spacing=self.h)
        psi_B = PS2AE(self.calc_B, grid_apspacing=self.h)

        ns = self.calc_A.wfs.nspins

        # weight functions sum_c VcWc
        wa = []
        wb = []

        for a in range(len(self.Va)):
            wa.append(
                interpolate_weight(self.calc_A, self.fineweightA[a], h=self.h))
        for b in range(len(self.Vb)):
            wb.append(
                interpolate_weight(self.calc_B, self.fineweightB[b], h=self.h))
        wa = np.asarray(wa)
        wb = np.asarray(wb)

        # check number of occupied and total number of bands
        n_occup_A = self.get_n_occupied_bands(self.calc_A)
        n_occup_B = self.get_n_occupied_bands(self.calc_B)

        # place to store <i_A(k)|w|j_B(k)>
        w_kij_AB = []
        w_kij_BA = []

        # k-point weights
        w_k = np.zeros(self.calc_A.wfs.kd.nibzkpts, dtype=np.float32)

        # get weight matrix at for each ij band at kpt and spin
        # the resulting matrix is organized in alpha and beta blocks
        #        |       |    |
        #        |  a    | 0  |    a:<psi_a|Vb*wb|psi_a> != 0
        # VW_AB =|_______|____|  , <psi_a|w|psi_b> = 0
        #        |   0   |  b |    b:<psi_b|Vb*wb|psi_b> != 0
        #        |       |    |
        #
        # a = nAa x nAa, b = nAb x nAb

        for spin in range(ns):
            for k in range(self.calc_A.wfs.kd.nibzkpts):

                # k-dependent overlap/pair density matrices
                inv_S = np.linalg.inv(self.n_ab[k])
                det_S = np.linalg.det(self.n_ab[k])
                I = np.identity(inv_S.shape[0])
                C_ab = np.transpose(np.dot(inv_S, (det_S * I)))

                nAa, nAb, nBa, nBb = self.check_bands(n_occup_A, n_occup_B, k)
                nas, n_occup, n_occup_s = self.check_spin_and_occupations(
                    nAa, nAb, nBa, nBb)

                # check that a and b cDFT states have similar spin state
                if np.sign(nAa - nAb) != np.sign(nBa - nBb):
                    warning = UserWarning(spin_state_error)
                    warnings.warn(warning)
                # form overlap matrices of correct size for each kpt

                if spin == 0:
                    w_kij_AB.append(
                        np.zeros((n_occup, n_occup), dtype=complex))
                    w_kij_BA.append(
                        np.zeros((n_occup, n_occup), dtype=complex))

                # store k-point weights
                kd = self.calc_A.wfs.kd
                w_kA = kd.weight_k[k]
                kd = self.calc_B.wfs.kd
                w_kB = kd.weight_k[k]

                for i in range(n_occup_s[spin]):
                    for j in range(n_occup_s[spin]):
                        I = spin * nas + i
                        J = spin * nas + j

                        psi_kA = psi_A.get_wave_function(n=i,
                                                         k=k,
                                                         s=spin,
                                                         ae=True)
                        psi_kB = psi_B.get_wave_function(n=j,
                                                         k=k,
                                                         s=spin,
                                                         ae=True)
                        w_ij_AB = []
                        w_ji_BA = []

                        for b in range(len(self.Vb)):
                            integral = psi_B.gd.integrate(
                                psi_kA.conj() * wb[b] * psi_kB,
                                global_integral=True) * C_ab[I][J]

                            if b >= self.n_charge_regionsB and spin == 1:
                                # for charge constraint w > 0
                                integral *= -1.
                            w_ij_AB.append(-integral)

                        for a in range(len(self.Va)):
                            integral = psi_A.gd.integrate(
                                psi_kB.conj() * wa[a] * psi_kA,
                                global_integral=True) * C_ab[J][I]
                            if a >= self.n_charge_regionsA and spin == 1:
                                integral *= -1.
                            w_ji_BA.append(-integral)

                        w_ij_AB = np.asarray(w_ij_AB) * Bohr**3
                        w_ji_BA = np.asarray(w_ji_BA) * Bohr**3

                        # collect kpt weight, only once per kpt
                        if spin == 0 and i == 0 and j == 0:
                            w_k[k] = (w_kA + w_kB) / 2.

                        w_kij_AB[k][I][J] += np.dot(self.Vb, w_ij_AB).sum()
                        w_kij_BA[k][J][I] += np.dot(self.Va, w_ji_BA).sum()

        self.w_k = w_k
        self.VW_AB = w_kij_AB
        self.VW_BA = w_kij_BA

        if self.save_matrix:
            np.save(self.VW_matrix + 'final_AB', self.VW_AB)
            np.save(self.VW_matrix + 'final_BA', self.VW_BA)
        return self.VW_AB, self.VW_BA, self.w_k

    def get_pair_weight_matrix(self):
        # <Psi_A|Psi_B> using pseudo wave functions and atomic corrections

        if not hasattr(self, 'n_ab'):
            raise ValueError(nab_missing_error)

        # check number of occupied and total number of bands
        n_occup_A = self.get_n_occupied_bands(
            self.calc_A)  # total of filled a and b bands
        n_occup_B = self.get_n_occupied_bands(self.calc_B)

        # place to store <i_A(k)|w|j_B(k)>
        w_kij_AB = []
        w_kij_BA = []
        # k-point weights
        w_k = np.zeros(self.calc_A.wfs.kd.nibzkpts)

        # get weight matrix at for each ij band at kpt and spin
        # the resulting matrix is organized in alpha and beta blocks
        #        |       |    |
        #        |  a    | 0  |    a:<psi_a|Vb*wb|psi_a> != 0
        # VW_AB =|_______|____|  , <psi_a|w|psi_b> = 0
        #        |   0   |  b |    b:<psi_b|Vb*wb|psi_b> != 0
        #        |       |    |
        #
        # a = nAa x nAa, b = nAb x nAb

        for kpt_a, kpt_b in zip(self.calc_A.wfs.kpt_u, self.calc_B.wfs.kpt_u):
            k = kpt_a.k
            spin = kpt_a.s

            # k-dependent overlap/pair density matrices
            inv_S = np.linalg.inv(self.n_ab[k])
            det_S = np.linalg.det(self.n_ab[k])
            I = np.identity(inv_S.shape[0])
            C_ab = np.transpose(np.dot(inv_S, (det_S * I)))

            inv_S = np.linalg.inv(np.transpose(self.n_ab[k]).conj())
            det_S = np.linalg.det(np.transpose(self.n_ab[k]))
            I = np.identity(inv_S.shape[0])
            C_ba = np.transpose(np.dot(inv_S, (det_S * I)))

            nAa, nAb, nBa, nBb = self.check_bands(n_occup_A, n_occup_B, k)
            nas, n_occup, n_occup_s = self.check_spin_and_occupations(
                nAa, nAb, nBa, nBb)

            # check that a and b cDFT states have similar spin state
            if np.sign(nAa - nAb) != np.sign(nBa - nBb):
                warning = UserWarning(spin_state_error)
                warnings.warn(warning)
            # form overlap matrices of correct size for each kpt

            if spin == 0:
                w_kij_AB.append(np.zeros((n_occup, n_occup), dtype=complex))
                w_kij_BA.append(np.zeros((n_occup, n_occup), dtype=complex))

            # store k-point weights
            kd = self.calc_A.wfs.kd
            w_kA = kd.weight_k[k]
            kd = self.calc_B.wfs.kd
            w_kB = kd.weight_k[k]

            for b in range(len(self.Vb)):
                self.get_matrix_element(kpt_a.psit_nG,
                                        kpt_a.P_ani,
                                        kpt_b.psit_nG,
                                        kpt_b.P_ani,
                                        n_occup_s,
                                        n_occup_s,
                                        w_kij_AB,
                                        self.regionsB[b],
                                        k,
                                        spin,
                                        nas,
                                        V=self.Vb[b],
                                        W=self.coarseweightB[b],
                                        C12=C_ab)

                if b >= self.n_charge_regionsB and spin == 1:
                    # change sign of b spin constraint
                    w_kij_AB[k][nas:, nas:] *= -1.

            for a in range(len(self.Va)):
                self.get_matrix_element(kpt_b.psit_nG,
                                        kpt_b.P_ani,
                                        kpt_a.psit_nG,
                                        kpt_a.P_ani,
                                        n_occup_s,
                                        n_occup_s,
                                        w_kij_BA,
                                        self.regionsA[a],
                                        k,
                                        spin,
                                        nas,
                                        V=self.Va[a],
                                        W=self.coarseweightA[a],
                                        C12=C_ba)
                if a >= self.n_charge_regionsA and spin == 1:
                    w_kij_BA[k][nas:, nas:] *= -1.

            if spin == 0:
                w_k[k] = (w_kA + w_kB) / 2.

        self.VW_BA = np.asarray(w_kij_BA)
        self.VW_AB = np.asarray(w_kij_AB)
        self.w_k = w_k

        return self.VW_AB, self.VW_BA, self.w_k

    def get_weight_matrix(self):
        # from the pair density matrix
        if not (hasattr(self, 'VW_AB')):
            if self.AE:
                self.get_ae_pair_weight_matrix()
            else:
                self.get_pair_weight_matrix()
        if not (hasattr(self, 'w_k')):
            self.get_ae_pair_density_matrix()

        if rank == 0:
            # diagonal of V*weight matrix
            self.VW = np.zeros((2, 2))
            # fill diagonal

            self.VW[0][0] += np.sum(self.NA)
            self.VW[1][1] += np.sum(self.NB)

            W_k_AB = np.zeros(len(self.w_k))
            W_k_BA = np.zeros(len(self.w_k))

            for k in range(len(self.w_k)):
                # sum_k (sum_ij <i|sum_c Vc*wc| j> * C_ij
                W_k_AB[k] = self.VW_AB[k].real.sum()
                W_k_AB[k] *= self.w_k[k]
                W_k_BA[k] = self.VW_BA[k].real.sum()
                W_k_BA[k] *= self.w_k[k]

            self.VW[0][1] = W_k_AB.sum()
            self.VW[1][0] = W_k_BA.sum()

            return self.VW

    def get_ae_pair_density_matrix(self, calc_A, calc_B, matrix_name=None):
        if calc_A.wfs.kd.nibzkpts != 1:
            raise ValueError(ae_ibz_error)
        # <Psi_A|Psi_B> using the all-electron pair density
        psi_A = PS2AE(calc_A, grid_spacing=self.h)
        psi_B = PS2AE(calc_B, grid_spacing=self.h)

        ns = calc_A.wfs.nspins

        # total of filled a and b bands for each spin and kpt
        n_occup_A = self.get_n_occupied_bands(calc_A)
        n_occup_B = self.get_n_occupied_bands(calc_B)

        # list to store k-dependent pair density
        n_AB = []
        w_k = np.zeros(calc_A.wfs.kd.nibzkpts)  # store kpt weights

        # get overlap at for each ij band at kpt and spin
        # the resulting matrix is organized in alpha and beta blocks
        #    |       |    |
        #    |  a    | 0  |    a:<psi_a|psi_a> != 0
        # S =|_______|____|  , <psi_a|psi_b> = 0
        #    |   0   |  b |    b:<psi_b|psi_b> != 0
        #    |       |    |
        #
        # a = naa x naa, b = nab x nab

        for spin in range(ns):
            for k in range(calc_A.wfs.kd.nibzkpts):

                nAa, nAb, nBa, nBb = self.check_bands(n_occup_A, n_occup_B, k)
                nas, n_occup, n_occup_s = self.check_spin_and_occupations(
                    nAa, nAb, nBa, nBb)
                kd = calc_A.wfs.kd
                w_kA = kd.weight_k[k]
                kd = calc_B.wfs.kd
                w_kB = kd.weight_k[k]

                # check that a and b cDFT states have similar spin state
                if np.sign(nAa - nAb) != np.sign(nBa - nBb):
                    warning = UserWarning(
                        'The cDFT wave functions have\n'
                        'different spin states! Similar\n'
                        'spin states are required for coupling constant\n'
                        'calculation!')
                    warnings.warn(warning)
                # form overlap matrices of correct size for each kpt
                if spin == 0:
                    n_AB.append(np.zeros((n_occup, n_occup), dtype=complex))

                for i in range(n_occup_s[spin]):
                    for j in range(n_occup_s[spin]):
                        # take only the bands which contain electrons
                        # in spin-orbital
                        psi_kA = psi_A.get_wave_function(n=i,
                                                         k=k,
                                                         s=spin,
                                                         ae=True)
                        psi_kB = psi_B.get_wave_function(n=j,
                                                         k=k,
                                                         s=spin,
                                                         ae=True)
                        n_ij = psi_A.gd.integrate(psi_kA.conj(),
                                                  psi_kB,
                                                  global_integral=True)

                        if spin == 0 and i == 0 and j == 0:
                            w_k[k] = (w_kA + w_kB) / 2.

                        I = spin * nas + i
                        J = spin * nas + j
                        n_AB[k][I][J] = n_ij * Bohr**3

        n_AB = np.asarray(n_AB)
        self.w_k = w_k
        self.n_ab = n_AB

        if self.save_matrix:
            if matrix_name is None:
                np.save(self.S_matrix + 'final', self.n_ab)
            else:
                np.save('%s_final' % matrix_name, self.n_ab)
        return self.n_ab, self.w_k

    def get_pair_density_matrix(self, calc_A, calc_B, matrix_name=None):
        # <Psi_A|Psi_B> using pseudo wave functions and atomic corrections
        assert calc_A.wfs.kd.nibzkpts == calc_B.wfs.kd.nibzkpts

        # total of filled a and b bands for each spin and kpt
        n_occup_A = self.get_n_occupied_bands(calc_A)
        n_occup_B = self.get_n_occupied_bands(calc_B)

        # list to store k-dependent pair density
        n_AB = []

        w_k = np.zeros(calc_A.wfs.kd.nibzkpts)  # store kpt weights

        # get overlap at for each ij band at kpt and spin
        # the resulting matrix is organized in alpha and beta blocks
        #    |       |    |
        #    |  a    | 0  |    a:<psi_a|psi_a> != 0
        # S =|_______|____|  , <psi_a|psi_b> = 0
        #    |   0   |  b |    b:<psi_b|psi_b> != 0
        #    |       |    |
        #
        # a = naa x naa, b = nab x nab

        for kpt_a, kpt_b in zip(calc_A.wfs.kpt_u, calc_B.wfs.kpt_u):

            k = kpt_a.k
            spin = kpt_a.s
            nAa, nAb, nBa, nBb = self.check_bands(n_occup_A, n_occup_B, k)
            nas, n_occup, n_occup_s = self.check_spin_and_occupations(
                nAa, nAb, nBa, nBb)

            # check that a and b cDFT states have similar spin state
            if np.sign(nAa - nAb) != np.sign(nBa - nBb):
                warning = UserWarning(
                    'The cDFT wave functions have\n'
                    'different spin states! Similar\n'
                    'spin states are required for coupling constant\n'
                    'calculation!')
                warnings.warn(warning)
            # form overlap matrices of correct size for each kpt
            if spin == 0:
                n_AB.append(np.zeros((n_occup, n_occup), dtype=complex))

            kd = calc_A.wfs.kd
            w_kA = kd.weight_k[k]
            kd = calc_B.wfs.kd
            w_kB = kd.weight_k[k]

            self.get_matrix_element(kpt_b.psit_nG, kpt_b.P_ani, kpt_a.psit_nG,
                                    kpt_a.P_ani, n_occup_s, n_occup_s, n_AB,
                                    None, k, spin, nas)

            if spin == 0:
                w_k[k] = (w_kA + w_kB) / 2.
        n_AB = np.asarray(n_AB)
        self.w_k = w_k
        self.n_ab = n_AB

        if self.save_matrix:
            if matrix_name is None:
                np.save(self.S_matrix + 'final', self.n_ab)
            else:
                np.save('%s_final' % matrix_name, self.n_ab)

        return self.n_ab, self.w_k

    def get_overlap_matrix(self, save=False, name='wf_overlap'):

        # from the pair density matrix
        if not (hasattr(self, 'n_ab')):
            if self.AE:
                self.n_ab, self.w_k = self.get_ae_pair_density_matrix(
                    self.calc_A, self.calc_B)
            else:
                self.n_ab, self.w_k = self.get_pair_density_matrix(
                    self.calc_A, self.calc_B)

        if not (hasattr(self, 'w_k')):
            self.get_ae_pair_density_matrix(self.calc_A, self.calc_B)
        if rank == 0:
            self.S = np.identity(2)

            S_k_AB = np.zeros(len(self.w_k))
            S_k_BA = np.zeros(len(self.w_k))

            for k in range(len(self.w_k)):
                S_k_AB[k] = self.w_k[k] * np.linalg.det(self.n_ab[k]).real
                # determinant of the complex conjugate
                S_k_BA[k] = self.w_k[k] * np.linalg.det(
                    np.transpose(self.n_ab[k]).conj()).real

            S_AB = S_k_AB.sum()
            S_BA = S_k_BA.sum()

            # fill 2x2 overlap matrix
            self.S[0][1] = S_AB
            self.S[1][0] = S_BA

            if save:
                np.save('%s' % name, self.S)
            return self.S

    def get_migliore_wf_overlaps(self, calc_A, calc_B, calc_gs):
        '''A = <I|GS>, B= <F|GS>, S = <I|F> '''
        if self.AE:
            S, w_AB = self.get_ae_pair_density_matrix(calc_A,
                                                      calc_B,
                                                      matrix_name='Sif')
            A, w_AC = self.get_ae_pair_density_matrix(calc_A,
                                                      calc_gs,
                                                      matrix_name='A')
            B, w_BC = self.get_ae_pair_density_matrix(calc_B,
                                                      calc_gs,
                                                      matrix_name='B')
        else:
            S, w_AB = self.get_pair_density_matrix(calc_A,
                                                   calc_B,
                                                   matrix_name='Sif')
            A, w_AC = self.get_pair_density_matrix(calc_A,
                                                   calc_gs,
                                                   matrix_name='A')
            B, w_BC = self.get_pair_density_matrix(calc_B,
                                                   calc_gs,
                                                   matrix_name='B')

        return S, A, B, w_AB, w_AC, w_BC

    def get_matrix_element(self,
                           psit1_nG,
                           P1_ani,
                           psit2_nG,
                           P2_ani,
                           n_occup_1,
                           n_occup_2,
                           vw,
                           region,
                           k,
                           s,
                           nas,
                           V=1.,
                           W=1.,
                           C12=None):
        # weight: V*W acting on psitb_nG: VW_ij = <psita_nG_i|VW|psitb_nG_j> or
        # overlap <psita_nG_i|psitb_nG_j>
        # includes occupation dependency and only states with filled
        # orbitals are considered
        # Cab = cofactor matrix
        # k and s --> kpt and spin

        VW_ij = np.zeros((len(psit1_nG), len(psit2_nG)), complex)
        VW = V * W

        for n1 in range(len(psit1_nG)):
            for n2 in range(len(psit2_nG)):
                nijt_G = np.multiply(psit1_nG[n1].conj(), psit2_nG[n2])
                VW_ij[n1][n2] = self.gd.integrate(VW * nijt_G,
                                                  global_integral=True)

        P_array = np.zeros(VW_ij.shape, complex)
        for a, P1_ni in P1_ani.items():
            P2_ni = P2_ani[a]
            if region is None or a in region:
                # the atomic correction is zero outside
                # the augmentation regions --> w=0 if
                # a is not in the w.
                # region = None --> overlap and all terms are used
                inner = np.dot(P1_ni, self.calc_A.wfs.setups[a].dO_ii)
                outer = (np.dot(P2_ni, V * np.conjugate(inner.T))).T
                P_array += outer

        self.calc_A.density.gd.comm.sum(P_array)
        VW_ij += P_array

        for i, j in enumerate(range(n_occup_1[s])):
            for x, y in enumerate(range(n_occup_2[s])):
                I = s * nas + i
                X = s * nas + x
                if C12 is None:
                    vw[k][I][X] += VW_ij[j][y]
                else:
                    vw[k][I][X] += VW_ij[j][y] * C12[I][X]

    def get_reorganization_energy(self):
        # get Ea (Rb) - Ea(Ra) -->
        # cdft energy at geometry Rb
        # with charge constraint A
        geometry = self.calc_B.get_atoms()

        cdft = self.cdft_A
        # set cdft_a on geometry of B
        geometry.calc = cdft
        self.reorg = geometry.get_potential_energy()
        self.reorg -= self.EA
        return self.reorg

    def get_energy_gap(self):
        # get Eb (Ra) - Ea(Ra) -->
        # cdft energy at geometry Ra
        # with charge constraint A or B
        geometry = self.calc_A.get_atoms()

        cdft = self.cdft_B
        # set cdft_a on geometry of B
        geometry.calc = cdft
        self.energy_gap = geometry.get_potential_energy()
        self.energy_gap -= self.EA
        return self.energy_gap

    def get_landau_zener(self):
        # computes the Landau-Zener factor

        planck = 4.135667e-15  # eV s

        if self.reorg is not None:
            Lambda = self.reorg
        else:
            Lambda = self.get_reorganization_energy()

        if hasattr(self, 'ct'):
            Hab = self.ct
        else:
            Hab = self.get_coupling_term()

        self.P_lz = 1 - np.exp(-np.pi**(3 / 2) * (np.abs(Hab))**2 /
                               (planck * self.freq *
                                np.sqrt(self.temp * Lambda * kb)))

        return self.P_lz

    def get_marcus_rate(self):
        # adiabatic transition state energy
        if self.E_tst:
            E_tst = self.E_tst
        else:
            # compute the activation energy
            # from the classical marcus
            # parabolas
            E_tst = self.get_marcus_barrier()

        # electron transmission coeffiecient
        # is the reaction diabatic or adiabatic?
        P_lz = self.get_landau_zener()
        dE = self.EA - self.EB
        if dE >= -self.reorg:
            # normal
            kappa = 2. * P_lz / (1 + P_lz)
        else:
            kappa = 2 * P_lz * (1 - P_lz)
        rate = kappa * self.freq * np.exp(-E_tst / (kb * self.temp))

        return rate

    def get_marcus_barrier(self):
        # approximate barrier from
        # two marcus parabolas
        # and an adiabatic correction

        # reaction energy
        dE = self.reaction_energy

        # crossing of the parabolas
        barrier = 1 / (4 * self.reorg) * (self.reorg + dE)**2
        if self.use_adiabatic_correction:
            # adiabatic correction
            correction = (
                np.abs(self.ct) +
                (self.reorg + dE) / 2. -
                np.sqrt(((self.reorg + self.dE)**2) / 4. +
                        (np.abs(self.ct))**2))

            return barrier - correction
        else:
            return barrier

    def get_n_occupied_bands(self, calc):
        ''' how many occupied bands?'''

        ns = calc.wfs.nspins
        occup_ks = np.zeros((len(calc.wfs.kd.weight_k), ns), dtype=int)
        w = calc.wfs.kd.weight_k
        for k in range(calc.wfs.kd.nibzkpts):
            for s in range(2):
                f_n = calc.get_occupation_numbers(kpt=k, spin=s)
                f_n *= 1. / w[k]
                f_N = f_n > self.band_occupation_cutoff
                occup_ks[k][s] += f_N.sum()
        return occup_ks

    def check_bands(self, n_occup_A, n_occup_B, k):
        if not self.specific_bands_A and not self.specific_bands_B:
            nAa, nAb = n_occup_A[k][0], n_occup_A[k][1]
            nBa, nBb = n_occup_B[k][0], n_occup_B[k][1]
        else:  # choose subset of orbitals for coupling
            if self.specific_bands_A:
                nAa, nAb = self.specific_bands_A[0], self.specific_bands_A[1]
            if self.specific_bands_B:
                nBa, nBb = self.specific_bands_B[0], self.specific_bands_B[1]

        return nAa, nAb, nBa, nBb

    def check_spin_and_occupations(self, nAa, nAb, nBa, nBb):
        nas = np.max((nAa, nBa))
        nbs = np.max((nAb, nBb))
        n_occup_s = [nas, nbs]
        n_occup = sum(n_occup_s)

        return nas, n_occup, n_occup_s



from math import pi

from ase.units import Ha
from gpaw.core import PWDesc, UGDesc
from gpaw.core.atom_arrays import AtomDistribution
from gpaw.core.domain import Domain
from gpaw.core.matrix import Matrix
from gpaw.core.plane_waves import PWArray
from gpaw.new import cached_property, zips
from gpaw.new.builder import create_uniform_grid
from gpaw.new.pw.hamiltonian import PWHamiltonian, SpinorPWHamiltonian
from gpaw.new.pw.poisson import make_poisson_solver
from gpaw.new.pw.pot_calc import PlaneWavePotentialCalculator
from gpaw.new.pwfd.builder import PWFDDFTComponentsBuilder
from gpaw.new.spinors import SpinorWaveFunctionDescriptor
from gpaw.new.xc import create_functional
from gpaw.typing import Array1D


class PWDFTComponentsBuilder(PWFDDFTComponentsBuilder):
    interpolation = 'fft'

    def __init__(self, atoms, params, *, comm, ecut=340, qspiral=None):
        self.ecut = ecut / Ha
        super().__init__(atoms, params, comm=comm)

        self.qspiral_v = (None if qspiral is None else
                          qspiral @ self.grid.icell * (2 * pi))
        self._nct_ag = None
        self._tauct_ag = None

    @cached_property
    def atomdist(self) -> AtomDistribution:
        # We should just distribute the atom evenly, but that is not compatible
        # with LCAO initialization!
        # return AtomDistribution.from_number_of_atoms(len(self.fracpos_ac),
        #                                              self.communicators['d'])
        return AtomDistribution(
            self.grid.ranks_from_fractional_positions(self.fracpos_ac),
            self.grid.comm)

    def create_uniform_grids(self):
        grid = create_uniform_grid(
            'pw',
            self.params.gpts,
            self.atoms.cell,
            self.atoms.pbc,
            self.ibz.symmetries,
            h=self.params.h,
            interpolation='fft',
            ecut=self.ecut,
            comm=self.communicators['d'])
        fine_grid = grid.new(size=grid.size_c * 2)
        # decomposition=[2 * d for d in grid.decomposition]
        return grid, fine_grid

    def create_wf_description(self) -> Domain:
        pw = PWDesc(ecut=self.ecut,
                    cell=self.grid.cell,
                    comm=self.grid.comm,
                    dtype=self.dtype)
        if self.ncomponents == 4:
            return SpinorWaveFunctionDescriptor(pw, qspiral_v=self.qspiral_v)
        return pw

    def create_xc_functional(self):
        if self.params.xc['name'] in ['HSE06', 'PBE0', 'EXX']:
            return ...
        return create_functional(self._xc,
                                 self.fine_grid)

    @cached_property
    def interpolation_desc(self):
        return PWDesc(ecut=2 * self.ecut,
                      cell=self.grid.cell,
                      comm=self.grid.comm)

    @cached_property
    def electrostatic_potential_desc(self):
        return self.interpolation_desc.new(ecut=8 * self.ecut)

    def get_pseudo_core_densities(self):
        if self._nct_ag is None:
            self._nct_ag = self.setups.create_pseudo_core_densities(
                self.interpolation_desc, self.fracpos_ac, self.atomdist,
                xp=self.xp)
        return self._nct_ag

    def get_pseudo_core_ked(self):
        if self._tauct_ag is None:
            self._tauct_ag = self.setups.create_pseudo_core_ked(
                self.interpolation_desc, self.fracpos_ac, self.atomdist)
        return self._tauct_ag

    def create_poisson_solver(self, fine_pw, params):
        return make_poisson_solver(fine_pw,
                                   self.fine_grid,
                                   self.atoms.pbc,
                                   self.params.charge,
                                   **params)

    def create_potential_calculator(self):
        nct_ag = self.get_pseudo_core_densities()
        pw = nct_ag.pw
        fine_pw = self.electrostatic_potential_desc
        poisson_solver = self.create_poisson_solver(
            fine_pw,
            self.params.poissonsolver or {'strength': 1.0})
        return PlaneWavePotentialCalculator(
            self.grid, self.fine_grid,
            pw, fine_pw,
            self.setups,
            self.xc,
            poisson_solver,
            fracpos_ac=self.fracpos_ac,
            atomdist=self.atomdist,
            soc=self.soc,
            xp=self.xp)

    def create_hamiltonian_operator(self, blocksize=10):
        if self.ncomponents < 4:
            return PWHamiltonian(self.grid, self.wf_desc, self.xp)
        return SpinorPWHamiltonian()

    def convert_wave_functions_from_uniform_grid(self,
                                                 C_nM: Matrix,
                                                 basis_set,
                                                 kpt_c,
                                                 q):
        # Replace this with code that goes directly from C_nM to
        # psit_nG via PWAtomCenteredFunctions.
        # XXX

        grid = self.grid.new(kpt=kpt_c, dtype=self.dtype)
        pw = self.wf_desc.new(kpt=kpt_c)
        psit_nG = pw.empty(self.nbands, self.communicators['b'])

        if self.dtype == complex:
            emikr_R = grid.eikr(-kpt_c)

        mynbands, M = C_nM.dist.shape
        if self.ncomponents < 4:
            psit_nR = grid.zeros(mynbands)
            basis_set.lcao_to_grid(C_nM.data, psit_nR.data, q)

            for psit_R, psit_G in zips(psit_nR, psit_nG, strict=False):
                if self.dtype == complex:
                    psit_R.data *= emikr_R
                psit_R.fft(out=psit_G)
        else:
            psit_sR = grid.empty(2)
            C_nsM = C_nM.data.reshape((mynbands, 2, M // 2))
            for psit_sG, C_sM in zips(psit_nG, C_nsM, strict=False):
                psit_sR.data[:] = 0.0
                basis_set.lcao_to_grid(C_sM, psit_sR.data, q)
                psit_sR.data *= emikr_R
                for psit_G, psit_R in zips(psit_sG, psit_sR):
                    psit_R.fft(out=psit_G)

        return psit_nG.to_xp(self.xp)

    def read_ibz_wave_functions(self, reader):
        ibzwfs = super().read_ibz_wave_functions(reader)

        if 'coefficients' not in reader.wave_functions:
            return ibzwfs

        c = reader.bohr**1.5
        if reader.version < 0:
            c = 1  # old gpw file
        elif reader.version < 4:
            c /= self.grid.size_c.prod()

        index_kG = reader.wave_functions.indices

        for wfs in ibzwfs:
            pw = self.wf_desc.new(kpt=wfs.kpt_c)
            if wfs.spin == 0:
                check_g_vector_ordering(self.grid, pw, index_kG[wfs.k])

            index = (wfs.spin, wfs.k) if self.ncomponents != 4 else (wfs.k,)
            data = reader.wave_functions.proxy('coefficients', *index)
            data.scale = c
            data.length_of_last_dimension = pw.shape[-1]

            if self.communicators['w'].size == 1:
                orig_shape = data.shape
                data.shape = (self.nbands, ) + pw.shape
                wfs.psit_nX = PWArray(pw, self.nbands, data=data)
                data.shape = orig_shape
            else:
                band_comm = self.communicators['b']
                wfs.psit_nX = PWArray(pw, self.nbands, comm=band_comm)
                if pw.comm.rank == 0:
                    mynbands = (self.nbands +
                                band_comm.size - 1) // band_comm.size
                    n1 = min(band_comm.rank * mynbands, self.nbands)
                    n2 = min((band_comm.rank + 1) * mynbands, self.nbands)
                    assert wfs.psit_nX.mydims[0] == n2 - n1
                    data = data[n1:n2]  # read from file
                for psit_G, array in zips(wfs.psit_nX, data):
                    psit_G.scatter_from(array)

        return ibzwfs


def check_g_vector_ordering(grid: UGDesc,
                            pw: PWDesc,
                            index_G: Array1D) -> None:
    size = tuple(grid.size)
    if pw.dtype == float:
        size = (size[0], size[1], size[2] // 2 + 1)
    index0_G = pw.indices(size)
    nG = len(index0_G)
    assert (index0_G == index_G[:nG]).all()
    assert (index_G[nG:] == -1).all()



"""Python wrapper for GPAW's native XC functionals."""

import _gpaw
from gpaw import debug

codes = {
    'LDA': -1,
    'PBE': 0,
    'revPBE': 1,
    'RPBE': 2,
    'PW91': 14,
    'TPSS': 20,
    'M06-L': 21,
    'revTPSS': 22}
# NOTE: when adding MGGA functionals to the above
# list, self.type must be set to MGGA in XCKernel:__init__


class XCNull:
    type = 'LDA'
    name = 'null'

    def calculate(self, e_g, n_sg, dedn_sg):
        e_g[:] = 0.0


class XCKernel:
    def __init__(self, name):
        self.name = name
        if name == 'LDA':
            self.type = 'LDA'
        elif name == 'TPSS' or name == 'M06-L' or name == 'revTPSS':
            self.type = 'MGGA'
        else:
            self.type = 'GGA'
        self.xc = _gpaw.XCFunctional(codes[name])

    def calculate(self, e_g, n_sg, dedn_sg,
                  sigma_xg=None, dedsigma_xg=None,
                  tau_sg=None, dedtau_sg=None):
        """Calculate energy and derivatives from density and gradients.

         * e_g is the energy density.  Values are overwritten.
         * n_sg is the density and is an input.
         * dedn_sg is the partial derivative of the energy with
           respect to the density (any gradients constant).  Values are
           added to this array.
         * sigma_xg is the squared norm of the gradient and is an input.
         * dedsigma_xg is the partial derivative of the energy with respect
           to the squared gradient norm.  Values are overwritten.
         * tau_sg and dedtau_sg probably behave similarly but correspond to
           2nd-order derivatives for MGGAs.  XXX verify and document this.
        """
        if debug:
            self.check_arguments(e_g, n_sg, dedn_sg, sigma_xg, dedsigma_xg,
                                 tau_sg, dedtau_sg)
        self.xc.calculate(e_g, n_sg, dedn_sg, sigma_xg, dedsigma_xg, tau_sg,
                          dedtau_sg)

    def check_arguments(self, e_g, n_sg, dedn_sg, sigma_xg, dedsigma_xg,
                        tau_sg, dedtau_sg):
        S = n_sg.shape[0]
        assert 1 <= S <= 2
        G = n_sg.shape[1:]
        assert e_g.shape == G
        assert e_g.flags.contiguous and e_g.dtype == float
        assert dedn_sg.shape == (S,) + G
        assert dedn_sg.flags.contiguous
        assert dedn_sg.dtype == float
        if self.type != 'LDA':
            assert sigma_xg.shape == (2 * S - 1,) + G
            assert dedsigma_xg.shape == (2 * S - 1,) + G
            assert sigma_xg.flags.contiguous and sigma_xg.dtype == float
            assert (dedsigma_xg.flags.contiguous and
                    dedsigma_xg.dtype == float)
            if self.type == 'MGGA':
                assert tau_sg.shape == (S,) + G
                assert dedtau_sg.shape == (S,) + G
                assert tau_sg.flags.contiguous and tau_sg.dtype == float
                assert (dedtau_sg.flags.contiguous and
                        dedtau_sg.dtype == float)



from ase import Atoms
from gpaw import GPAW
from gpaw.test import equal
from gpaw.xc.tools import vxc


def test_xc_degeneracy():
    a = 5.0
    d = 1.0
    x = d / 3**0.5
    atoms = Atoms('CH4',
                  [(0.0, 0.0, 0.0),
                   (x, x, x),
                   (-x, -x, x),
                   (x, -x, -x),
                   (-x, x, -x)],
                  cell=(a, a, a),
                  pbc=False)

    atoms.positions[:] += a / 2
    params = dict(
        mode='fd', h=0.25, nbands=4, convergence={'eigenstates': 7.8e-10})
    atoms.calc = GPAW(**params)
    energy = atoms.get_potential_energy()

    # The three eigenvalues e[1], e[2], and e[3] must be degenerate:
    e = atoms.calc.get_eigenvalues()
    print(e[1] - e[3])
    equal(e[1], e[3], 9.3e-8)

    energy_tolerance = 0.002
    equal(energy, -23.631, energy_tolerance)

    gs = atoms.calc.gs_adapter()

    # Calculate non-selfconsistent PBE eigenvalues:
    epbe0 = e[:2] - vxc(gs, n2=2)[0, 0] + vxc(gs, 'PBE', n2=2)[0, 0]

    # Calculate selfconsistent PBE eigenvalues:
    atoms.calc = GPAW(**params, xc='PBE')
    energy = atoms.get_potential_energy()
    epbe = atoms.calc.get_eigenvalues()

    de = epbe[1] - epbe[0]
    de0 = epbe0[1] - epbe0[0]
    print(de, de0)
    equal(de, de0, 0.001)



import numpy as np
import scipy.sparse as sparse
from ase.neighborlist import PrimitiveNeighborList
# from ase.utils.timing import timer
from gpaw.utilities.tools import tri2full

# from gpaw import debug
from gpaw.lcao.overlap import (FourierTransformer, TwoSiteOverlapCalculator,
                               ManySiteOverlapCalculator,
                               AtomicDisplacement, NullPhases, BlochPhases,
                               DerivativeAtomicDisplacement)


def get_cutoffs(f_Ij):
    rcutmax_I = []
    for f_j in f_Ij:
        rcutmax = 0.001  # 'paranoid zero'
        for f in f_j:
            rcutmax = max(rcutmax, f.get_cutoff())
        rcutmax_I.append(rcutmax)
    return rcutmax_I


def get_lvalues(f_Ij):
    return [[f.get_angular_momentum_number() for f in f_j] for f_j in f_Ij]


class AtomPairRegistry:
    def __init__(self, cutoff_a, pbc_c, cell_cv, spos_ac):
        nl = PrimitiveNeighborList(cutoff_a, skin=0, sorted=True,
                                   self_interaction=True,
                                   use_scaled_positions=True)

        nl.update(pbc=pbc_c, cell=cell_cv, coordinates=spos_ac)
        r_and_offset_aao = {}

        def add(a1, a2, R_c, offset):
            r_and_offset_aao.setdefault((a1, a2), []).append((R_c, offset))

        for a1, spos1_c in enumerate(spos_ac):
            a2_a, offsets = nl.get_neighbors(a1)
            for a2, offset in zip(a2_a, offsets):
                spos2_c = spos_ac[a2] + offset

                R_c = np.dot(spos2_c - spos1_c, cell_cv)
                add(a1, a2, R_c, offset)
                if a1 != a2 or offset.any():
                    add(a2, a1, -R_c, -offset)
        self.r_and_offset_aao = r_and_offset_aao

    def get(self, a1, a2):
        R_c_and_offset_a = self.r_and_offset_aao.get((a1, a2))
        return R_c_and_offset_a

    def get_atompairs(self):
        return list(sorted(self.r_and_offset_aao))


class TCIExpansions:
    def __init__(self, phit_Ij, pt_Ij, I_a):
        assert len(pt_Ij) == len(phit_Ij)

        # Cutoffs by species:
        pt_rcmax_I = get_cutoffs(pt_Ij)
        phit_rcmax_I = get_cutoffs(phit_Ij)
        rcmax_I = [max(rc1, rc2) for rc1, rc2
                   in zip(pt_rcmax_I, phit_rcmax_I)]

        transformer = FourierTransformer(rcut=max(rcmax_I + [1e-3]), N=2**10)
        tsoc = TwoSiteOverlapCalculator(transformer)
        msoc = ManySiteOverlapCalculator(tsoc, I_a, I_a)
        phit_Ijq = msoc.transform(phit_Ij)
        pt_Ijq = msoc.transform(pt_Ij)
        pt_l_Ij = get_lvalues(pt_Ij)
        phit_l_Ij = get_lvalues(phit_Ij)
        self.O_expansions = msoc.calculate_expansions(phit_l_Ij, phit_Ijq,
                                                      phit_l_Ij, phit_Ijq)
        self.T_expansions = msoc.calculate_kinetic_expansions(phit_l_Ij,
                                                              phit_Ijq)
        self.P_expansions = msoc.calculate_expansions(pt_l_Ij, pt_Ijq,
                                                      phit_l_Ij, phit_Ijq)
        self.I_a = I_a  # Actually I_a belongs outside, like spos_ac.
        self.rcmax_I = rcmax_I
        self.phit_rcmax_I = phit_rcmax_I
        self.pt_rcmax_I = pt_rcmax_I

    @classmethod
    def new_from_setups(cls, setups):
        I_setup = {}
        setups_I = list(setups.setups.values())
        for I, setup in enumerate(setups_I):
            I_setup[setup] = I
        I_a = [I_setup[setup] for setup in setups]

        return TCIExpansions([s.basis_functions_J for s in setups_I],
                             [s.pt_j for s in setups_I],
                             I_a)

    def get_tci_calculator(self, cell_cv, spos_ac, pbc_c, ibzk_qc, dtype):
        return TCICalculator(self, cell_cv, spos_ac, pbc_c, ibzk_qc, dtype)

    def get_manytci_calculator(self, setups, gd, spos_ac, ibzk_qc, dtype,
                               timer):
        return ManyTCICalculator(self, setups, gd, spos_ac, ibzk_qc, dtype,
                                 timer)


class TCICalculator:
    """High-level two-center integral calculator.

    This object is not aware of parallelization.  It works with any
    pair of atoms a1, a2.

    Create the object and calculate any interatomic overlap matrix as below.

      tci = TCI(...)

    Projector/basis overlap <pt_i^a1|phi_mu> between atoms a1, a2:

      P_qim = tci.P(a1, a2)

    Derivatives of the above with respect to movement of a2:

      dPdR_qvim = tci.dPdR(a1, a2)

    Basis/basis overlap and kinetic matrix elements between atoms a1, a2:

      O_qmm, T_qmm = tci.O_T(a1, a2)

    Derivative of the above wrt. position of a2:

      dOdR_qvmm, dTdR_qvmm = tci.dOdR_dTdR(a1, a2)

    """
    def __init__(self, tciexpansions, cell_cv, spos_ac, pbc_c, ibzk_qc,
                 dtype):

        self.tciexpansions = tciexpansions
        self.dtype = dtype

        # XXX It is somewhat nasty that rcmax depends on how long our
        # longest orbital happens to be
        # Cutoffs by atom:
        I_a = tciexpansions.I_a
        cutoff_a = [tciexpansions.rcmax_I[I] for I in I_a]
        self.pt_rcmax_a = np.array([tciexpansions.pt_rcmax_I[I] for I in I_a])
        self.phit_rcmax_a = np.array([tciexpansions.phit_rcmax_I[I]
                                      for I in I_a])

        self.a1a2 = AtomPairRegistry(cutoff_a, pbc_c, cell_cv, spos_ac)

        self.ibzk_qc = ibzk_qc
        if ibzk_qc.any():
            self.get_phases = BlochPhases
        else:
            self.get_phases = NullPhases

        self.O_T = self._tci_shortcut(False, False)
        self.P = self._tci_shortcut(True, False)
        self.dOdR_dTdR = self._tci_shortcut(False, True)
        self.dPdR = self._tci_shortcut(True, True)

    def _tci_shortcut(self, P, derivative):
        def calculate(a1, a2):
            return self._calculate(a1, a2, P, derivative)
        return calculate

    def _calculate(self, a1, a2, P=False, derivative=False):
        """Calculate overlap of functions between atoms a1 and a2."""

        # We want to see quickly if there is no overlap because distance
        # outside bounding spheres.

        R_c_and_offset_a = self.a1a2.get(a1, a2)
        if R_c_and_offset_a is None:
            return None if P else (None, None)

        rcut1 = self.pt_rcmax_a[a1] if P else self.phit_rcmax_a[a1]
        rcut2 = self.phit_rcmax_a[a2]
        maxdist = rcut1 + rcut2

        # Filter out displacements larger than maxdist:
        R_c_and_offset_a = [obj for obj in R_c_and_offset_a
                            if np.linalg.norm(obj[0]) < maxdist]
        if not R_c_and_offset_a:  # There was no overlap after all
            return None if P else (None, None)

        dtype = self.dtype
        get_phases = self.get_phases

        displacement = (DerivativeAtomicDisplacement
                        if derivative
                        else AtomicDisplacement)
        ibzk_qc = self.ibzk_qc
        nq = len(ibzk_qc)
        phit_rcmax_a = self.phit_rcmax_a
        pt_rcmax_a = self.pt_rcmax_a

        shape = (nq, 3) if derivative else (nq,)

        if P:
            P_expansion = self.tciexpansions.P_expansions.get(a1, a2)
            obj = P_qim = P_expansion.zeros(shape, dtype=dtype)
        else:
            O_expansion = self.tciexpansions.O_expansions.get(a1, a2)
            T_expansion = self.tciexpansions.T_expansions.get(a1, a2)
            O_qmm = O_expansion.zeros(shape, dtype=dtype)
            T_qmm = T_expansion.zeros(shape, dtype=dtype)
            obj = O_qmm, T_qmm

        for R_c, offset in R_c_and_offset_a:
            norm = np.linalg.norm(R_c)
            phases = get_phases(ibzk_qc, offset)

            disp = displacement(None, a1, a2, R_c, offset, phases)

            if P:
                assert norm < pt_rcmax_a[a1] + phit_rcmax_a[a2]
                disp.evaluate_overlap(P_expansion, P_qim)
            else:
                assert norm < phit_rcmax_a[a1] + phit_rcmax_a[a2]
                disp.evaluate_overlap(O_expansion, O_qmm)
                disp.evaluate_overlap(T_expansion, T_qmm)

        return obj


class ManyTCICalculator:
    def __init__(self, tciexpansions, setups, gd, spos_ac, ibzk_qc, dtype,
                 timer):
        self.tci = tciexpansions.get_tci_calculator(gd.cell_cv, spos_ac,
                                                    gd.pbc_c,
                                                    ibzk_qc, dtype)

        self.setups = setups
        self.dtype = dtype
        self.Pindices = setups.projector_indices()
        self.Mindices = setups.basis_indices()
        self.natoms = len(setups)
        self.nq = len(ibzk_qc)
        self.nao = self.Mindices.max
        self.timer = timer

    # @timer('tci-projectors')
    def P_aqMi(self, my_atom_indices, derivative=False):
        P_axMi = {}
        if derivative:
            P = self.tci.dPdR

            def empty(nI):
                return np.empty((self.nq, 3, self.nao, nI), self.dtype)
        else:
            P = self.tci.P

            def empty(nI):
                return np.empty((self.nq, self.nao, nI), self.dtype)

        Mindices = self.Mindices

        for a1 in my_atom_indices:
            P_xMi = empty(self.setups[a1].ni)

            for a2 in range(self.natoms):
                N1, N2 = Mindices[a2]
                P_xmi = P_xMi[..., N1:N2, :]
                P_xim = P(a1, a2)
                if P_xim is None:
                    P_xmi[:] = 0.0
                else:
                    P_xmi[:] = P_xim.swapaxes(-2, -1).conj()
            P_axMi[a1] = P_xMi

        if derivative:
            for a in P_axMi:
                P_axMi[a] *= -1.0
        return P_axMi

    # @timer('tci-sparseprojectors')
    def P_qIM(self, my_atom_indices):
        nq = self.nq
        P = self.tci.P
        P_qIM = [sparse.dok_matrix((self.Pindices.max, self.Mindices.max),
                                   dtype=self.dtype)
                 for _ in range(nq)]

        for a1 in my_atom_indices:
            I1, I2 = self.Pindices[a1]

            # We can stride a2 over e.g. bd.comm and then do bd.comm.sum().
            # How should we do comm.sum() on a sparse matrix though?
            for a2 in range(self.natoms):
                M1, M2 = self.Mindices[a2]
                P_qim = P(a1, a2)
                if P_qim is not None:
                    for q in range(nq):
                        P_qIM[q][I1:I2, M1:M2] = P_qim[q]
        P_qIM = [P_IM.tocsr() for P_IM in P_qIM]
        return P_qIM

    # @timer('tci-bfs')
    def O_qMM_T_qMM(self, gdcomm, Mstart, Mstop, ignore_upper=False,
                    derivative=False):
        mynao = Mstop - Mstart
        Mindices = self.Mindices

        if derivative:
            O_T = self.tci.dOdR_dTdR
            shape = (self.nq, 3, mynao, self.nao)
        else:
            O_T = self.tci.O_T
            shape = (self.nq, mynao, self.nao)

        O_xMM = np.zeros(shape, self.dtype)
        T_xMM = np.zeros(shape, self.dtype)

        # XXX the a1/a2 loops are not yet well load balanced.
        for a1 in range(self.natoms):
            M1, M2 = Mindices[a1]
            if M2 <= Mstart or M1 >= Mstop:
                continue

            myM1 = max(M1 - Mstart, 0)
            myM2 = min(M2 - Mstart, mynao)
            nM = myM2 - myM1

            assert nM > 0, nM

            a2max = a1 + 1  # if not derivative else self.natoms

            for a2 in range(gdcomm.rank, a2max, gdcomm.size):
                O_xmm, T_xmm = O_T(a1, a2)
                if O_xmm is None:
                    continue

                N1, N2 = Mindices[a2]
                m1 = max(Mstart - M1, 0)
                m2 = m1 + nM  # (Slice may go beyond end of matrix but OK)
                O_xmm = O_xmm[..., m1:m2, :]
                T_xmm = T_xmm[..., m1:m2, :]
                O_xMM[..., myM1:myM2, N1:N2] = O_xmm
                T_xMM[..., myM1:myM2, N1:N2] = T_xmm

        if not ignore_upper and O_xMM.size:  # reshape() fails on size-0 arrays
            assert mynao == self.nao
            assert O_xMM.shape[-2:] == (self.nao, self.nao)
            if derivative:
                def lumap(arr, out):
                    np.conj(arr, out)
                    out *= -1.0
            else:
                lumap = np.conj

            for arr_xMM in [O_xMM, T_xMM]:
                for tmp_MM in arr_xMM.reshape(-1, self.nao, self.nao):
                    tri2full(tmp_MM, UL='L', map=lumap)

        return O_xMM, T_xMM



# this test should coverage the save and restore of
# fermi-levels when using fixmagmom:
#
# yes, fermi-level-splitting sounds a little bit strange

from ase import Atoms
from gpaw import GPAW, FermiDirac, MixerSum
from gpaw.test import equal


def test_fermisplit(in_tmp_dir):
    calc = GPAW(mode='fd',
                occupations=FermiDirac(width=0.1, fixmagmom=True),
                mixer=MixerSum(beta=0.05, nmaxold=3, weight=50.0),
                convergence={'energy': 0.1, 'eigenstates': 1.5e-1,
                             'density': 1.5e-1})
    atoms = Atoms('Cr', pbc=False)
    atoms.center(vacuum=4)
    mm = [1] * 1
    mm[0] = 6.
    atoms.set_initial_magnetic_moments(mm)
    atoms.calc = calc
    atoms.get_potential_energy()

    ef1 = calc.get_fermi_levels().mean()
    efsplit1 = calc.get_fermi_levels().ptp()

    ef3 = calc.get_fermi_levels()
    calc.write('test.gpw')

    # check number one: is the splitting value saved?
    readtest = GPAW('test.gpw')
    ef2 = readtest.get_fermi_levels().mean()
    efsplit2 = readtest.get_fermi_levels().ptp()

    # numpy arrays
    ef4 = readtest.get_fermi_levels()

    # These values should be identic
    equal(ef1, ef2, 1e-9)
    equal(efsplit1, efsplit2, 1e-9)
    equal(ef3.mean(), ef1, 1e-9)
    equal(ef3.mean(), ef2, 1e-9)
    equal(ef3.mean(), ef4.mean(), 1e-9)
    equal(ef3[0] - ef3[1], ef4[0] - ef4[1], 1e-9)
    equal(efsplit1, ef4[0] - ef4[1], 1e-9)



"""
Calculates Raman matrices from Raman tensor
"""
import numpy as np
from typing import Tuple

from ase.units import invcm
from ase.utils.filecache import MultiFileJSONCache
from gpaw.typing import ArrayND


def gaussian(w, sigma):
    g = 1.0 / (np.sqrt(2.0 * np.pi) * sigma) * np.exp(-(w**2) / (2 * sigma**2))
    return g


class RamanData:
    def __init__(self, raman_name="Rlab", gridspacing=1.0) -> None:
        """Raman Spectroscopy data.

        Parameters
        ----------
        raman_name: str
            Name of Rlab file cache. Default 'Rlab'
        gridspacing: float
            grid spacing in cm^-1, default 1 rcm
        """
        # JSON cache
        self.raman_cache = MultiFileJSONCache(raman_name)
        self.wph_w = self.raman_cache["phonon_frequencies"]  # eV
        self.wrl_w = self.raman_cache["frequency_grid"]  # eV

        gridmax = self.wph_w[-1] + 6.3e-3  # highest freq + 50rcm
        self.gridev_w = np.linspace(
            0.0,
            gridmax,
            num=int(gridmax / (gridspacing * invcm) + 1),
        )  # eV

    def calculate_raman_intensity(self, d_i, d_o, T=300, sigma=5.0):
        """
        Calculates Raman intensities from Raman tensor.

        Returns bare `|R|^2` and and Bose occupation weighted values

        Parameters
        ----------
        d_i: int
            Incoming polarization
        d_o: int
            Outgoing polarization
        T: float
            Temperature in Kelvin
        sigma: float
            Gaussian line width in rcm, default 5rcm
        """

        KtoeV = 8.617278e-5
        sigmaev = sigma * invcm  # 1eV = 8.065544E3 rcm

        # grid for spectrum with extra space and sigma/5 spacing
        if self.gridev_w is None:
            self.gridev_w = np.linspace(
                0.0,
                self.wph_w[-1] * 1.1,
                num=int(self.wph_w[-1] / (sigmaev / 5) + 100),
            )

        int_bare = np.zeros_like(self.gridev_w, dtype=float)
        int_occ = np.zeros_like(self.gridev_w, dtype=float)

        # Note: Resonant Raman does not have a frequency grid
        R_lw = self.raman_cache[f"{'xyz'[d_i]}{'xyz'[d_o]}"]

        for l in range(len(self.wph_w)):
            occ = 1.0 / (np.exp(self.wph_w[l] / (KtoeV * T)) - 1.0) + 1.0
            delta_w = gaussian((self.gridev_w - self.wph_w[l]), sigmaev)
            R2_w = np.abs(R_lw[l]).real
            R2d_w = R2_w * delta_w
            int_bare += R2d_w
            int_occ += occ / self.wph_w[l] * R2d_w

        return int_bare, int_occ

    def calculate_raman_spectrum(self,
                                 entries, T=300, sigma=5.0
                                 ) -> Tuple[ArrayND, ArrayND]:
        """
        Calculates Raman intensities from Raman tensor.

        Returns Raman shift in eV, bare `|R|^2`
        and Bose occupation weighted `|R|^2` values

        Parameters
        ----------
        entries: str, list
            Sting or list of strings with desired polarisaitons
            For example: ["xx", "yy", "xy", "yx"]
        T: float
            Temperature in Kelvin
        sigma: float
            Gaussian line width in rcm, default 5rcm
        """
        if isinstance(entries, str):
            entries = [
                entries,
            ]

        spectrum_w = np.zeros_like(self.gridev_w)

        polnum = {"x": 0, "y": 1, "z": 2}
        for entry in entries:
            d_i = polnum[entry[0]]
            d_o = polnum[entry[1]]
            (int_bare, _) = self.calculate_raman_intensity(d_i, d_o, T, sigma)
            spectrum_w += int_bare

        return self.gridev_w, spectrum_w

    @classmethod
    def plot_raman(
        cls,
        figname,
        grid_w,
        spectra_nw,
        labels_n=None,
        relative=False,
    ):
        """Plots a given Raman spectrum.

        Parameters
        ----------
        figname: str
            Filename for figure.
        grid_w:
            Frequency grid of spectrum in eV
        spectra_nw: np.ndarray
            Raman spectra to plot
        labels_n: list
            Labels for the legend
        relative: bool
            If true, normalize each spectrum to 1
            Default is False
        """

        from scipy import signal
        import matplotlib.pyplot as plt

        if not isinstance(spectra_nw, np.ndarray):
            spectra_nw = np.asarray(spectra_nw)

        ylabel = "Intensity (arb. units)"
        if relative:
            ylabel = "I/I_max"
            spectra_nw /= np.max(spectra_nw, axis=1)[:, np.newaxis]
        else:
            spectra_nw /= np.max(spectra_nw)

        nspec = spectra_nw.shape[0]
        if labels_n is None:
            labels_n = nspec * [None]

        grid_w = cls.eVtorcm(grid_w)

        for i in range(nspec):
            peaks = signal.find_peaks(spectra_nw[i])[0]
            locations = np.take(grid_w, peaks)
            intensities = np.take(spectra_nw[i], peaks)

            plt.plot(grid_w, spectra_nw[i], label=labels_n[i])

            for j, loc in enumerate(locations):
                if intensities[j] / np.max(intensities) > 0.05:
                    plt.axvline(x=loc, color="grey", linestyle="--")

        plt.minorticks_on()
        if labels_n[0] is not None:
            plt.legend()
        plt.title("Raman intensity")
        plt.xlabel("Raman shift (cm$^{-1}$)")
        plt.ylabel(ylabel)
        if relative:
            plt.yticks([])

        plt.savefig(figname, dpi=300)
        plt.clf()

    @classmethod
    def eVtorcm(cls, energy):
        return energy / invcm



import time


class QuadraticETA:
    def __init__(self, min_times=10):
        self.times = []
        self.min_times = max(min_times, 3)

    def update(self):
        self.times.append(time.time())

    def eta(self, n):
        if (len(self.times) < self.min_times):
            # print '%6d' % (len(self.times))
            return 0.0

        ns = int(len(self.times) / 3.)

        y1 = 0.0
        for i in range(len(self.times) - 3 * ns, len(self.times) - 2 * ns):
            y1 += self.times[i]
        y1 /= ns
        x1 = (len(self.times) - 3 * ns + len(self.times) - 2 * ns) / 2.

        y2 = 0.0
        for i in range(len(self.times) - 2 * ns, len(self.times) - ns):
            y2 += self.times[i]
        y2 /= ns
        x2 = (len(self.times) - 2 * ns + len(self.times) - 1 * ns) / 2.

        y3 = 0.0
        for i in range(len(self.times) - ns, len(self.times)):
            y3 += self.times[i]
        y3 /= ns
        x3 = (len(self.times) - 1 * ns + len(self.times) - 0 * ns) / 2.

        # 2nd order fit:
        #   y(x) = a * x**2 + b * x + c
        #   y'(x) = 2a * x + b
        #   y''(x) = 2a
        #   y2''(x2) = 2 a = [(y3-y2)/(x3-x2) -
        #                     (y2-y1)/(x2-x1)] / [.5*(x3+x2)-.5*(x2+x1)]
        #   y2'(x2) = 2*a*x + b = (y3-y1)/(x3-x1)
        #   y2(x2) = a * x2**2 + b * x2 + c
        a = .5 * ((y3 - y2) / (x3 - x2) - (y2 - y1) /
                  (x2 - x1)) / (.5 * (x3 + x2) - .5 * (x2 + x1))
        b = (y3 - y1) / (x3 - x1) - 2 * a * x2
        c = y2 - a * x2 * x2 - b * x2

        # print x1,y1
        # print x2,y2
        # print x3,y3

        # print a
        # print b
        # print c

        return a * n * n + b * n + c - time.time()


if __name__ == "__main__":
    eta = QuadraticETA(10)

    N = 100

    print(time.time())
    t0 = time.time()
    for i in range(N):
        eta.update()
        eta.eta(N + 1)
        for j in range(i + 1):
            time.sleep(.001)

    t1 = time.time()

    print(time.time())

    print(t1 - t0)



# Copyright (C) 2006 CSC-Scientific Computing Ltd.
# Please see the accompanying LICENSE file for further information.
import os
import sys
import re
import shlex
from sysconfig import get_config_var, get_platform
from subprocess import run
from pathlib import Path
from stat import ST_MTIME


def config_args(key):
    return shlex.split(get_config_var(key))


def mtime(path, name, mtimes):
    """Return modification time.

    The modification time of a source file is returned.  If one of its
    dependencies is newer, the mtime of that file is returned.
    This function fails if two include files with the same name
    are present in different directories."""

    include = re.compile(r'^#\s*include "(\S+)"', re.MULTILINE)

    if name in mtimes:
        return mtimes[name]
    t = os.stat(os.path.join(path, name))[ST_MTIME]
    for name2 in include.findall(open(os.path.join(path, name)).read()):
        path2, name22 = os.path.split(name2)
        if name22 != name:
            t = max(t, mtime(os.path.join(path, path2), name22, mtimes))
    mtimes[name] = t
    return t


def check_dependencies(sources):
    # Distutils does not do deep dependencies correctly.  We take care of
    # that here so that "python setup.py build_ext" always does the right
    # thing!
    mtimes = {}  # modification times

    # Remove object files if any dependencies have changed:
    plat = get_platform() + '-{maj}.{min}'.format(maj=sys.version_info[0],
                                                  min=sys.version_info[1])
    remove = False
    for source in sources:
        path, name = os.path.split(source)
        t = mtime(path + '/', name, mtimes)
        o = 'build/temp.%s/%s.o' % (plat, source[:-2])  # object file
        if os.path.exists(o) and t > os.stat(o)[ST_MTIME]:
            print('removing', o)
            os.remove(o)
            remove = True

    so = 'build/lib.{}/_gpaw.so'.format(plat)
    if os.path.exists(so) and remove:
        # Remove shared object C-extension:
        # print 'removing', so
        os.remove(so)


def write_configuration(define_macros, include_dirs, libraries, library_dirs,
                        extra_link_args, extra_compile_args,
                        runtime_library_dirs, extra_objects, compiler):

    # Write the compilation configuration into a file
    try:
        out = open('configuration.log', 'w')
    except IOError as x:
        print(x)
        return
    print("Current configuration", file=out)
    print("compiler", compiler, file=out)
    print("libraries", libraries, file=out)
    print("library_dirs", library_dirs, file=out)
    print("include_dirs", include_dirs, file=out)
    print("define_macros", define_macros, file=out)
    print("extra_link_args", extra_link_args, file=out)
    print("extra_compile_args", extra_compile_args, file=out)
    print("runtime_library_dirs", runtime_library_dirs, file=out)
    print("extra_objects", extra_objects, file=out)
    out.close()


def build_interpreter(
        compiler, extension, extension_objects, *,
        build_temp, build_bin, debug):
    exename = compiler.executable_filename('gpaw-python')
    print(f'building {repr(exename)} executable', flush=True)

    macros = extension.define_macros.copy()
    for undef in extension.undef_macros:
        macros.append((undef,))

    # Compile the sources that define GPAW_INTERPRETER
    sources = ['c/main.c']
    objects = compiler.compile(
        sources,
        output_dir=str(build_temp),
        macros=macros,
        include_dirs=extension.include_dirs,
        debug=debug,
        extra_postargs=extension.extra_compile_args)
    objects += extension_objects

    # Note: LDFLAGS and LIBS go together, but depending on the platform,
    # it might be unnecessary to include them
    extra_preargs = config_args('LDFLAGS')
    extra_postargs = (config_args('BLDLIBRARY')
                      + config_args('LIBS')
                      + config_args('LIBM')
                      + extension.extra_link_args
                      + config_args('LINKFORSHARED'))

    # Link the custom interpreter
    compiler.link_executable(
        objects, exename,
        output_dir=str(build_bin),
        extra_preargs=extra_preargs,
        libraries=extension.libraries,
        library_dirs=extension.library_dirs,
        runtime_library_dirs=extension.runtime_library_dirs,
        extra_postargs=extra_postargs,
        debug=debug,
        target_lang=extension.language)
    return build_bin / exename


def build_gpu(gpu_compiler, gpu_compile_args, gpu_include_dirs,
              define_macros, undef_macros, build_temp):
    print('building gpu kernels', flush=True)

    kernels_dpath = Path('c/gpu/kernels')

    # Create temp build directory
    build_temp_kernels_dpath = build_temp / kernels_dpath
    if not build_temp_kernels_dpath.exists():
        print(f'creating {build_temp_kernels_dpath}', flush=True)
        build_temp_kernels_dpath.mkdir(parents=True)

    # Glob all kernel files, but remove those included by other kernels
    kernels = sorted(kernels_dpath.glob('*.cpp'))
    for name in ['lfc-reduce.cpp',
                 'lfc-reduce-kernel.cpp',
                 'reduce.cpp',
                 'reduce-kernel.cpp']:
        kernels.remove(kernels_dpath / name)

    # Compile GPU kernels
    objects = []
    for src in kernels:
        obj = build_temp / src.with_suffix('.o')
        objects.append(str(obj))
        run_args = [gpu_compiler]
        run_args += gpu_compile_args
        for (name, value) in define_macros:
            arg = f'-D{name}'
            if value is not None:
                arg += f'={value}'
            run_args += [arg]
        run_args += [f'-U{name}' for name in undef_macros]
        run_args += [f'-I{dpath}' for dpath in gpu_include_dirs]
        run_args += ['-c', str(src)]
        run_args += ['-o', str(obj)]
        print(shlex.join(run_args), flush=True)
        p = run(run_args, check=False, shell=False)
        if p.returncode != 0:
            print(f'error: command {repr(gpu_compiler)} failed '
                  f'with exit code {p.returncode}',
                  file=sys.stderr, flush=True)
            sys.exit(1)

    return objects



import pytest
from ase import Atom, Atoms
from ase.parallel import parprint

from gpaw.test import equal
from gpaw import GPAW, mpi
from gpaw.lrtddft import LrTDDFT


@pytest.mark.lrtddft
@pytest.mark.libxc
def test_lrtddft_apmb():
    txt = '-'
    txt = '/dev/null'

    load = False

    if not load:
        R = 0.7  # approx. experimental bond length
        a = 3.0
        c = 4.0
        H2 = Atoms([Atom('H', (a / 2, a / 2, (c - R) / 2)),
                    Atom('H', (a / 2, a / 2, (c + R) / 2))],
                   cell=(a, a, c))
        calc = GPAW(mode='fd', xc='PBE', nbands=2, spinpol=False, txt=txt)
        H2.calc = calc
        H2.get_potential_energy()
    else:
        calc = GPAW('H2.gpw', txt=txt)

    # DFT only

    xc = 'LDA'

    # no spin

    lr = LrTDDFT(calc, xc=xc)
    lr.diagonalize()

    lr_ApmB = LrTDDFT(calc, xc=xc, force_ApmB=True)
    lr_ApmB.diagonalize()
    parprint('lr=', lr)
    parprint('ApmB=', lr_ApmB)
    equal(lr[0].get_energy(), lr_ApmB[0].get_energy(), 5.e-9)

    # with spin
    parprint('------ with spin')

    if not load:
        c_spin = GPAW(mode='fd', xc='PBE', nbands=2,
                      spinpol=True, parallel={'domain': mpi.world.size},
                      txt=txt)
        H2.calc = c_spin
        c_spin.calculate(H2)
    else:
        c_spin = GPAW('H2spin.gpw', txt=txt)
    lr = LrTDDFT(c_spin, xc=xc)
    lr.diagonalize()

    lr_ApmB = LrTDDFT(c_spin, xc=xc, force_ApmB=True)
    lr_ApmB.diagonalize()
    parprint('lr=', lr)
    parprint('ApmB=', lr_ApmB)
    equal(lr[0].get_energy(), lr_ApmB[0].get_energy(), 5.e-8)
    equal(lr[1].get_energy(), lr_ApmB[1].get_energy(), 5.e-8)

    # with spin virtual
    parprint('------ with virtual spin')

    lr = LrTDDFT(calc, xc=xc, nspins=2)
    lr.diagonalize()

    # ApmB
    lr_ApmB = LrTDDFT(calc, xc=xc, nspins=2)
    lr_ApmB.diagonalize()
    parprint('lr=', lr)
    parprint('ApmB=', lr_ApmB)
    equal(lr[0].get_energy(), lr_ApmB[0].get_energy(), 5.e-8)
    equal(lr[1].get_energy(), lr_ApmB[1].get_energy(), 5.e-8)

    # with HF exchange

    xc = 'PBE0'

    parprint('------ with spin xc=', xc)
    lr_spin = LrTDDFT(c_spin, xc=xc)
    lr_spin.diagonalize()
    parprint('lr=', lr_spin)

    parprint('------ with virtual spin xc=', xc)
    lr = LrTDDFT(calc, xc=xc, nspins=2)
    lr.diagonalize()
    parprint('lr=', lr)
    equal(lr[0].get_energy(), lr_spin[0].get_energy(), 3.8e-6)
    equal(lr[1].get_energy(), lr_spin[1].get_energy(), 3.4e-6)



import sys
from math import pi

import numpy as np
from ase.units import Hartree, Bohr

import gpaw.mpi as mpi

from gpaw.response.coulomb_kernels import CoulombKernel
from gpaw.response.density_kernels import get_density_xc_kernel
from gpaw.response.chi0 import Chi0Calculator, new_frequency_descriptor
from gpaw.response.pair import get_gs_and_context, KPointPairFactory


class DielectricFunctionCalculator:
    def __init__(self, chi0calc, truncation):
        from gpaw.response.pw_parallelization import Blocks1D
        self.chi0calc = chi0calc

        self.coulomb = CoulombKernel.from_gs(self.gs, truncation=truncation)
        self.context = chi0calc.context
        self.wd = chi0calc.wd
        self.blocks1d = Blocks1D(self.context.comm, len(self.wd))

        self._chi0cache = {}

    @property
    def gs(self):
        return self.chi0calc.gs

    def calculate_chi0(self, q_c, spin='all'):
        """Calculates the response function.

        Calculate the response function for a specific momentum.

        q_c: [float, float, float]
            The momentum wavevector.
        spin : str or int
            If 'all' then include all spins.
            If 0 or 1, only include this specific spin.
            (not used in transverse reponse functions)
        """

        # We cache the computed data since chi0 may otherwise be redundantly
        # calculated e.g. if the user calculates multiple directions.
        #
        # May be called multiple times with same q_c, and we want to
        # be able to recognize previous seen values of q_c.
        # We do this by rounding and converting to string with fixed
        # precision (so not very elegant).
        q_key = [f'{q:.10f}' for q in q_c]
        key = (spin, *q_key)

        # Spin='all' is a terrible cache key since it's inconsistent
        # with specifying spins one integer at the time.
        # We might as well change it to do the caching by spin index,
        # or maybe we can work around the caching entirely with a more
        # explicit API design.

        if key not in self._chi0cache:
            # We assume that the caller will trigger this multiple
            # times with the same qpoint, then several times with
            # another qpoint, etc.  If that's true, then we
            # need to cache no more than one qpoint at a time.
            # Thus to save memory, we clear the cache here.
            #
            # This should be replaced with something more reliable,
            # such as having the caller manage things more explicitly.
            #
            # See https://gitlab.com/gpaw/gpaw/-/issues/662
            #
            # In conclusion, delete the cache now:
            self._chi0cache.clear()

            chi0 = self.chi0calc.calculate(q_c, spin)
            chi0_wGG = chi0.body.get_distributed_frequencies_array()
            self.context.write_timer()
            things = chi0.qpd, chi0_wGG, chi0.chi0_WxvG, chi0.chi0_Wvv
            self._chi0cache[key] = things

        qpd, *more_things = self._chi0cache[key]
        return (qpd, *[thing.copy() if thing is not None else thing
                       for thing in more_things])

    def collect(self, a_w):
        return self.blocks1d.all_gather(a_w)

    def get_frequencies(self):
        """ Return frequencies that Chi is evaluated on"""
        return self.wd.omega_w * Hartree

    def get_chi(self, xc='RPA', q_c=[0, 0, 0], spin='all',
                direction='x', return_VchiV=True, q_v=None,
                rshelmax=-1, rshewmin=None):
        """ Returns v^1/2 chi v^1/2 for the density response and chi for the
        spin response. The truncated Coulomb interaction is included as
        v^-1/2 v_t v^-1/2. This is in order to conform with
        the head and wings of chi0, which is treated specially for q=0.

        spin : str or int
            If 'all' then include all spins.
            If 0 or 1, only include this specific spin.
            (not used in transverse reponse functions)
        rshelmax : int or None
            Expand kernel in real spherical harmonics inside augmentation
            spheres. If None, the kernel will be calculated without
            augmentation. The value of rshelmax indicates the maximum index l
            to perform the expansion in (l < 6).
        rshewmin : float or None
            If None, the kernel correction will be fully expanded up to the
            chosen lmax. Given as a float, (0 < rshewmin < 1) indicates what
            coefficients to use in the expansion. If any coefficient
            contributes with less than a fraction of rshewmin on average,
            it will not be included.
        """
        qpd, chi0_wGG, chi0_WxvG, chi0_Wvv = self.calculate_chi0(q_c, spin)

        coulomb_bare = CoulombKernel.from_gs(self.gs, truncation=None)
        Kbare_G = coulomb_bare.V(qpd=qpd, q_v=q_v)
        sqrtV_G = Kbare_G**0.5

        nG = len(sqrtV_G)

        Ktrunc_G = self.coulomb.V(qpd=qpd, q_v=q_v)

        if self.coulomb.truncation is None:
            K_GG = np.eye(nG, dtype=complex)
        else:
            K_GG = np.diag(Ktrunc_G / Kbare_G)

        if qpd.kd.gamma:
            if isinstance(direction, str):
                d_v = {'x': [1, 0, 0],
                       'y': [0, 1, 0],
                       'z': [0, 0, 1]}[direction]
            else:
                d_v = direction
            d_v = np.asarray(d_v) / np.linalg.norm(d_v)
            W = self.blocks1d.myslice
            chi0_wGG[:, 0] = np.dot(d_v, chi0_WxvG[W, 0])
            chi0_wGG[:, :, 0] = np.dot(d_v, chi0_WxvG[W, 1])
            chi0_wGG[:, 0, 0] = np.dot(d_v, np.dot(chi0_Wvv[W], d_v).T)

        if xc != 'RPA':
            Kxc_GG = get_density_xc_kernel(qpd,
                                           self.gs, self.context,
                                           functional=xc,
                                           chi0_wGG=chi0_wGG)
            K_GG += Kxc_GG / sqrtV_G / sqrtV_G[:, np.newaxis]

        # Invert Dyson eq.
        chi_wGG = []
        for chi0_GG in chi0_wGG:
            """v^1/2 chi0 V^1/2"""
            chi0_GG[:] = chi0_GG * sqrtV_G * sqrtV_G[:, np.newaxis]
            chi_GG = np.dot(np.linalg.inv(np.eye(nG) -
                                          np.dot(chi0_GG, K_GG)),
                            chi0_GG)
            if not return_VchiV:
                chi0_GG /= sqrtV_G * sqrtV_G[:, np.newaxis]
                chi_GG /= sqrtV_G * sqrtV_G[:, np.newaxis]
            chi_wGG.append(chi_GG)

        if len(chi_wGG):
            chi_wGG = np.array(chi_wGG)
        else:
            chi_wGG = np.zeros((0, nG, nG), complex)

        return qpd, chi0_wGG, np.array(chi_wGG)

    def get_dynamic_susceptibility(self, xc='ALDA', q_c=[0, 0, 0],
                                   q_v=None,
                                   rshelmax=-1, rshewmin=None,
                                   filename='chiM_w.csv'):
        """Calculate the dynamic susceptibility.

        Returns macroscopic(could be generalized?) dynamic susceptibility:
        chiM0_w, chiM_w = DielectricFunction.get_dynamic_susceptibility()
        """

        qpd, chi0_wGG, chi_wGG = self.get_chi(xc=xc, q_c=q_c,
                                              rshelmax=rshelmax,
                                              rshewmin=rshewmin,
                                              return_VchiV=False)

        rf0_w = np.zeros(len(chi_wGG), dtype=complex)
        rf_w = np.zeros(len(chi_wGG), dtype=complex)

        for w, (chi0_GG, chi_GG) in enumerate(zip(chi0_wGG, chi_wGG)):
            rf0_w[w] = chi0_GG[0, 0]
            rf_w[w] = chi_GG[0, 0]

        rf0_w = self.collect(rf0_w)
        rf_w = self.collect(rf_w)

        if filename is not None and mpi.rank == 0:
            write_response_function(filename, self.wd.omega_w * Hartree,
                                    rf0_w, rf_w)

        return rf0_w, rf_w

    def get_dielectric_matrix(self, xc='RPA', q_c=[0, 0, 0],
                              direction='x', symmetric=True,
                              calculate_chi=False, q_v=None,
                              add_intraband=False):
        r"""Returns the symmetrized dielectric matrix.

        ::

            \tilde\epsilon_GG' = v^{-1/2}_G \epsilon_GG' v^{1/2}_G',

        where::

            epsilon_GG' = 1 - v_G * P_GG' and P_GG'

        is the polarization.

        ::

            In RPA:   P = chi^0
            In TDDFT: P = (1 - chi^0 * f_xc)^{-1} chi^0

        in addition to RPA one can use the kernels, ALDA, Bootstrap and
        LRalpha (long-range kerne), where alpha is a user specified parameter
        (for example xc='LR0.25')

        The head of the inverse symmetrized dielectric matrix is equal
        to the head of the inverse dielectric matrix (inverse dielectric
        function)"""

        qpd, chi0_wGG, chi0_WxvG, chi0_Wvv = self.calculate_chi0(q_c)

        if add_intraband:
            print('add_intraband=True is not supported at this time')
            raise NotImplementedError

        K_G = self.coulomb.sqrtV(qpd=qpd, q_v=q_v)
        nG = len(K_G)

        if qpd.kd.gamma:
            if isinstance(direction, str):
                d_v = {'x': [1, 0, 0],
                       'y': [0, 1, 0],
                       'z': [0, 0, 1]}[direction]
            else:
                d_v = direction

            d_v = np.asarray(d_v) / np.linalg.norm(d_v)
            W = self.blocks1d.myslice
            chi0_wGG[:, 0] = np.dot(d_v, chi0_WxvG[W, 0])
            chi0_wGG[:, :, 0] = np.dot(d_v, chi0_WxvG[W, 1])
            chi0_wGG[:, 0, 0] = np.dot(d_v, np.dot(chi0_Wvv[W], d_v).T)
            if q_v is not None:
                print('Restoring q dependence of head and wings of chi0')
                chi0_wGG[:, 1:, 0] *= np.dot(q_v, d_v)
                chi0_wGG[:, 0, 1:] *= np.dot(q_v, d_v)
                chi0_wGG[:, 0, 0] *= np.dot(q_v, d_v)**2

        if xc != 'RPA':
            Kxc_GG = get_density_xc_kernel(qpd,
                                           self.gs, self.context,
                                           functional=xc,
                                           chi0_wGG=chi0_wGG)

        if calculate_chi:
            chi_wGG = []

        for chi0_GG in chi0_wGG:
            if xc == 'RPA':
                P_GG = chi0_GG
            else:
                P_GG = np.dot(np.linalg.inv(np.eye(nG) -
                                            np.dot(chi0_GG, Kxc_GG)),
                              chi0_GG)
            if symmetric:
                e_GG = np.eye(nG) - P_GG * K_G * K_G[:, np.newaxis]
            else:
                K_GG = (K_G**2 * np.ones([nG, nG])).T
                e_GG = np.eye(nG) - P_GG * K_GG
            if calculate_chi:
                K_GG = np.diag(K_G**2)
                if xc != 'RPA':
                    K_GG += Kxc_GG
                chi_wGG.append(np.dot(np.linalg.inv(np.eye(nG) -
                                                    np.dot(chi0_GG, K_GG)),
                                      chi0_GG))
            chi0_GG[:] = e_GG

        # chi0_wGG is now the dielectric matrix
        if calculate_chi:
            if len(chi_wGG):
                chi_wGG = np.array(chi_wGG)
            else:
                chi_wGG = np.zeros((0, nG, nG), complex)

        if not calculate_chi:
            return chi0_wGG
        else:
            # chi_wGG is the full density response function..
            return qpd, chi0_wGG, chi_wGG

    def get_dielectric_function(self, xc='RPA', q_c=[0, 0, 0], q_v=None,
                                direction='x', filename='df.csv'):
        """Calculate the dielectric function.

        Returns dielectric function without and with local field correction:
        df_NLFC_w, df_LFC_w = DielectricFunction.get_dielectric_function()
        """
        e_wGG = self.get_dielectric_matrix(xc, q_c, direction, q_v=q_v)
        df_NLFC_w = np.zeros(len(e_wGG), dtype=complex)
        df_LFC_w = np.zeros(len(e_wGG), dtype=complex)

        for w, e_GG in enumerate(e_wGG):
            df_NLFC_w[w] = e_GG[0, 0]
            df_LFC_w[w] = 1 / np.linalg.inv(e_GG)[0, 0]

        df_NLFC_w = self.collect(df_NLFC_w)
        df_LFC_w = self.collect(df_LFC_w)

        if filename is not None and mpi.rank == 0:
            write_response_function(filename, self.wd.omega_w * Hartree,
                                    df_NLFC_w, df_LFC_w)

        return df_NLFC_w, df_LFC_w

    def get_macroscopic_dielectric_constant(self, xc='RPA',
                                            direction='x', q_v=None):
        """Calculate macroscopic dielectric constant.

        Returns eM_NLFC and eM_LFC.

        Macroscopic dielectric constant is defined as the real part
        of dielectric function at w=0.

        Parameters:

        eM_LFC: float
            Dielectric constant without local field correction. (RPA, ALDA)
        eM2_NLFC: float
            Dielectric constant with local field correction.
        """

        self.context.print('', flush=False)
        self.context.print('%s Macroscopic Dielectric Constant:' % xc)

        df_NLFC_w, df_LFC_w = self.get_dielectric_function(
            xc=xc,
            filename=None,
            direction=direction,
            q_v=q_v)
        eps0 = np.real(df_NLFC_w[0])
        eps = np.real(df_LFC_w[0])
        self.context.print('  %s direction' % direction, flush=False)
        self.context.print('    Without local field: %f' % eps0, flush=False)
        self.context.print('    Include local field: %f' % eps)

        return eps0, eps

    def get_eels_spectrum(self, xc='RPA', q_c=[0, 0, 0],
                          direction='x', filename='eels.csv'):
        r"""Calculate EELS spectrum. By default, generate a file 'eels.csv'.

        EELS spectrum is obtained from the imaginary part of the
        density response function as, EELS(\omega) = - 4 * \pi / q^2 Im \chi.
        Returns EELS spectrum without and with local field corrections:

        df_NLFC_w, df_LFC_w = DielectricFunction.get_eels_spectrum()
        """

        # Calculate V^1/2 \chi V^1/2
        qpd, Vchi0_wGG, Vchi_wGG = self.get_chi(xc=xc, q_c=q_c,
                                                direction=direction)

        # Calculate eels = -Im 4 \pi / q^2  \chi
        eels_NLFC_w = -(1. / (1. - Vchi0_wGG[:, 0, 0])).imag
        eels_LFC_w = - (Vchi_wGG[:, 0, 0]).imag

        # Collect frequencies
        eels_NLFC_w = self.collect(eels_NLFC_w)
        eels_LFC_w = self.collect(eels_LFC_w)

        # Write to file
        if filename is not None and self.context.comm.rank == 0:
            omega_w = self.wd.omega_w
            write_response_function(filename, omega_w * Hartree,
                                    eels_NLFC_w, eels_LFC_w)

        return eels_NLFC_w, eels_LFC_w

    def get_polarizability(self, xc='RPA', direction='x', q_c=[0, 0, 0],
                           filename='polarizability.csv'):
        r"""Calculate the polarizability alpha.
        In 3D the imaginary part of the polarizability is related to the
        dielectric function by Im(eps_M) = 4 pi * Im(alpha). In systems
        with reduced dimensionality the converged value of alpha is
        independent of the cell volume. This is not the case for eps_M,
        which is ill-defined. A truncated Coulomb kernel will always give
        eps_M = 1.0, whereas the polarizability maintains its structure.

        By default, generate a file 'polarizability.csv'. The five columns are:
        frequency (eV), Real(alpha0), Imag(alpha0), Real(alpha), Imag(alpha)
        alpha0 is the result without local field effects and the
        dimension of alpha is \AA to the power of non-periodic directions
        """

        cell_cv = self.gs.gd.cell_cv
        pbc_c = self.gs.pbc

        if pbc_c.all():
            V = 1.0
        else:
            V = np.abs(np.linalg.det(cell_cv[~pbc_c][:, ~pbc_c]))

        if not self.coulomb.truncation:
            """Standard expression for the polarizability"""
            df0_w, df_w = self.get_dielectric_function(xc=xc,
                                                       q_c=q_c,
                                                       filename=None,
                                                       direction=direction)
            alpha_w = V * (df_w - 1.0) / (4 * pi)
            alpha0_w = V * (df0_w - 1.0) / (4 * pi)
        else:
            # Since eps_M = 1.0 for a truncated Coulomb interaction, it does
            # not make sense to apply it here. Instead one should define the
            # polarizability by
            #
            #     alpha * eps_M^{-1} = -L / (4 * pi) * <v_ind>
            #
            # where <v_ind> = 4 * pi * \chi / q^2 is the averaged induced
            # potential (relative to the strength of the  external potential).
            # With the bare Coulomb potential, this expression is equivalent to
            # the standard one. In a 2D system \chi should be calculated with a
            # truncated Coulomb potential and eps_M = 1.0

            self.context.print('Using truncated Coulomb interaction')

            qpd, chi0_wGG, chi_wGG = self.get_chi(xc=xc,
                                                  q_c=q_c,
                                                  direction=direction)
            alpha_w = -V * (chi_wGG[:, 0, 0]) / (4 * pi)
            alpha0_w = -V * (chi0_wGG[:, 0, 0]) / (4 * pi)

            alpha_w = self.collect(alpha_w)
            alpha0_w = self.collect(alpha0_w)

        # Convert to external units
        hypervol = Bohr**(sum(~pbc_c))
        alpha0_w *= hypervol
        alpha_w *= hypervol

        # Write results file
        if filename is not None and self.context.comm.rank == 0:
            omega_w = self.wd.omega_w
            write_response_function(filename, omega_w * Hartree,
                                    alpha0_w, alpha_w)

        return alpha0_w, alpha_w

    def check_sum_rule(self, spectrum=None):
        """Check f-sum rule.

        It takes the y of a spectrum as an entry and it checks its integral.

        spectrum: np.ndarray
            Input spectrum

        Note: not tested for spin response
        """

        assert (self.wd.omega_w[1:] - self.wd.omega_w[:-1]).ptp() < 1e-10

        if spectrum is None:
            raise ValueError('No spectrum input ')
        dw = self.wd.omega_w[1] - self.wd.omega_w[0]
        N1 = 0
        for iw in range(len(spectrum)):
            w = iw * dw
            N1 += spectrum[iw] * w
        N1 *= dw * self.gs.volume / (2 * pi**2)

        self.context.print('', flush=False)
        self.context.print('Sum rule:', flush=False)
        nv = self.gs.nvalence
        self.context.print('N1 = %f, %f  %% error' %
                           (N1, (N1 - nv) / nv * 100))


class DielectricFunction(DielectricFunctionCalculator):
    """This class defines dielectric function related physical quantities."""

    def __init__(self, calc, *,
                 frequencies=None,
                 domega0=None,  # deprecated
                 omega2=None,  # deprecated
                 omegamax=None,  # deprecated
                 ecut=50,
                 hilbert=True,
                 nbands=None, eta=0.2,
                 intraband=True, nblocks=1, world=mpi.world, txt=sys.stdout,
                 truncation=None, disable_point_group=False,
                 disable_time_reversal=False,
                 integrationmode=None, rate=0.0,
                 eshift=0.0):
        """Creates a DielectricFunction object.

        calc: str
            The ground-state calculation file that the linear response
            calculation is based on.
        frequencies:
            Input parameters for frequency_grid.
            Can be an array of frequencies to evaluate the response function at
            or dictionary of parameters for build-in nonlinear grid
            (see :ref:`frequency grid`).
        ecut: float
            Plane-wave cut-off.
        hilbert: bool
            Use hilbert transform.
        nbands: int
            Number of bands from calculation.
        eta: float
            Broadening parameter.
        intraband: bool
            Include intraband transitions.
        world: comm
            mpi communicator.
        nblocks: int
            Split matrices in nblocks blocks and distribute them G-vectors or
            frequencies over processes.
        txt: str
            Output file.
        truncation: str or None
            None for no truncation.
            '2D' for standard analytical truncation scheme.
            Non-periodic directions are determined from k-point grid
        eshift: float
            Shift unoccupied bands
        """

        gs, context = get_gs_and_context(calc, txt, world, timer=None)
        nbands = nbands or gs.bd.nbands

        wd = new_frequency_descriptor(gs, context, nbands, frequencies,
                                      domega0=domega0,
                                      omega2=omega2, omegamax=omegamax)

        kptpair_factory = KPointPairFactory(
            gs=gs, context=context, nblocks=nblocks)

        chi0calc = Chi0Calculator(
            wd=wd,
            kptpair_factory=kptpair_factory,
            ecut=ecut, nbands=nbands, eta=eta,
            hilbert=hilbert,
            intraband=intraband,
            disable_point_group=disable_point_group,
            disable_time_reversal=disable_time_reversal,
            integrationmode=integrationmode,
            rate=rate, eshift=eshift
        )

        super().__init__(chi0calc=chi0calc, truncation=truncation)


def write_response_function(filename, omega_w, rf0_w, rf_w):
    with open(filename, 'w') as fd:
        for omega, rf0, rf in zip(omega_w, rf0_w, rf_w):
            if rf0_w.dtype == complex:
                print('%.6f, %.6f, %.6f, %.6f, %.6f' %
                      (omega, rf0.real, rf0.imag, rf.real, rf.imag),
                      file=fd)
            else:
                print('%.6f, %.6f, %.6f' % (omega, rf0, rf), file=fd)


def read_response_function(filename):
    """Read a stored response function file"""
    d = np.loadtxt(filename, delimiter=',')
    omega_w = np.array(d[:, 0], float)

    if d.shape[1] == 3:
        # Real response function
        rf0_w = np.array(d[:, 1], float)
        rf_w = np.array(d[:, 2], float)
    elif d.shape[1] == 5:
        rf0_w = np.array(d[:, 1], complex)
        rf0_w.imag = d[:, 2]
        rf_w = np.array(d[:, 3], complex)
        rf_w.imag = d[:, 4]
    else:
        raise ValueError(f'Unexpected array dimension {d.shape}')

    return omega_w, rf0_w, rf_w
